import glob
import os
from collections.abc import Iterator
from contextlib import contextmanager
from typing import Callable, Optional

import click

from dagster import (
    DagsterInvariantViolationError,
    __version__ as dagster_version,
)
from dagster._cli.utils import (
    assert_no_remaining_opts,
    get_instance_for_cli,
    validate_dagster_home_is_set,
    validate_repo_has_defined_schedules,
)
from dagster._cli.workspace.cli_target import (
    RepositoryOpts,
    WorkspaceOpts,
    get_repository_from_cli_opts,
    repository_options,
    workspace_options,
)
from dagster._core.definitions.run_request import InstigatorType
from dagster._core.instance import DagsterInstance
from dagster._core.remote_representation import RemoteRepository
from dagster._core.scheduler.instigation import InstigatorStatus
from dagster._core.scheduler.scheduler import DagsterDaemonScheduler
from dagster._utils import PrintFn


@click.group(name="schedule")
def schedule_cli():
    """Commands for working with Dagster schedules."""


@schedule_cli.command(
    name="preview", help="Preview changes that will be performed by `dagster schedule up`."
)
@workspace_options
@repository_options
def schedule_preview_command(**opts: object):
    workspace_opts = WorkspaceOpts.extract_from_cli_options(opts)
    repository_opts = RepositoryOpts.extract_from_cli_options(opts)
    assert_no_remaining_opts(opts)
    return execute_preview_command(workspace_opts, repository_opts, click.echo)


def execute_preview_command(
    workspace_opts: WorkspaceOpts, repository_opts: RepositoryOpts, print_fn: PrintFn
):
    with (
        get_instance_for_cli() as instance,
        _get_repo(workspace_opts, repository_opts, instance) as repo,
    ):
        print_changes(repo, instance, print_fn, preview=True)


@schedule_cli.command(
    name="list",
    help="List all schedules that correspond to a repository.",
)
@click.option(
    "--running", "running_filter", help="Filter for running schedules", is_flag=True, default=False
)
@click.option(
    "--stopped", "stopped_filter", help="Filter for stopped schedules", is_flag=True, default=False
)
@click.option(
    "--name",
    "name_filter",
    help="Only display schedule schedule names",
    is_flag=True,
    default=False,
)
@workspace_options
@repository_options
def schedule_list_command(
    running_filter: bool, stopped_filter: bool, name_filter: bool, **other_opts: object
):
    workspace_opts = WorkspaceOpts.extract_from_cli_options(other_opts)
    repository_opts = RepositoryOpts.extract_from_cli_options(other_opts)
    assert_no_remaining_opts(other_opts)
    return execute_list_command(
        running_filter=running_filter,
        stopped_filter=stopped_filter,
        name_filter=name_filter,
        workspace_opts=workspace_opts,
        repository_opts=repository_opts,
        print_fn=click.echo,
    )


def execute_list_command(
    *,
    running_filter: bool,
    stopped_filter: bool,
    name_filter: bool,
    workspace_opts: WorkspaceOpts,
    repository_opts: RepositoryOpts,
    print_fn: PrintFn,
):
    with (
        get_instance_for_cli() as instance,
        _get_repo(workspace_opts, repository_opts, instance) as repo,
    ):
        if not name_filter:
            title = f"Repository {repo.name}"
            print_fn(title)
            print_fn("*" * len(title))

        repo_schedules = repo.get_schedules()
        stored_schedules_by_origin_id = {
            stored_schedule_state.instigator_origin_id: stored_schedule_state
            for stored_schedule_state in instance.all_instigator_state(
                repo.get_remote_origin_id(),
                repo.selector_id,
                instigator_type=InstigatorType.SCHEDULE,
            )
        }

        first = True

        for schedule in repo_schedules:
            schedule_state = schedule.get_current_instigator_state(
                stored_schedules_by_origin_id.get(schedule.get_remote_origin_id())
            )

            if running_filter and not schedule_state.is_running:
                continue
            if stopped_filter and schedule_state.is_running:
                continue

            if name_filter:
                print_fn(schedule.name)
                continue

            status = "RUNNING" if schedule_state.is_running else "STOPPED"
            schedule_title = f"Schedule: {schedule.name} [{status}]"
            if not first:
                print_fn("*" * len(schedule_title))

            first = False

            print_fn(schedule_title)
            print_fn(f"Cron Schedule: {schedule.cron_schedule}")


@schedule_cli.command(name="start", help="Start an existing schedule.")
@click.argument("schedule_name", required=False)
@click.option("--start-all", help="start all schedules", is_flag=True, default=False)
@workspace_options
@repository_options
def schedule_start_command(schedule_name: Optional[str], start_all: bool, **other_opts: object):
    workspace_opts = WorkspaceOpts.extract_from_cli_options(other_opts)
    repository_opts = RepositoryOpts.extract_from_cli_options(other_opts)
    assert_no_remaining_opts(other_opts)
    if schedule_name is None and start_all is False:
        print(  # noqa: T201
            "Noop: dagster schedule start was called without any arguments specifying which "
            "schedules to start. Pass a schedule name or the --start-all flag to start schedules."
        )
        return
    return execute_start_command(
        schedule_name=schedule_name,
        start_all=start_all,
        workspace_opts=workspace_opts,
        repository_opts=repository_opts,
        print_fn=click.echo,
    )


def execute_start_command(
    *,
    schedule_name: Optional[str],
    start_all: bool,
    workspace_opts: WorkspaceOpts,
    repository_opts: RepositoryOpts,
    print_fn: PrintFn,
):
    with (
        get_instance_for_cli() as instance,
        _get_repo(workspace_opts, repository_opts, instance) as repo,
    ):
        if start_all:
            for remote_schedule in repo.get_schedules():
                try:
                    instance.start_schedule(remote_schedule)
                except DagsterInvariantViolationError as ex:
                    raise click.UsageError(ex)  # pyright: ignore[reportArgumentType]

            print_fn(f"Started all schedules for repository {repo.name}")
        else:
            if not schedule_name:
                raise click.UsageError("Missing schedule name")
            try:
                instance.start_schedule(repo.get_schedule(schedule_name))
            except DagsterInvariantViolationError as ex:
                raise click.UsageError(ex)  # pyright: ignore[reportArgumentType]

            print_fn(f"Started schedule {schedule_name}")


@schedule_cli.command(name="stop", help="Stop an existing schedule.")
@click.argument("schedule_name")
@workspace_options
@repository_options
def schedule_stop_command(schedule_name: str, **other_opts: object):
    workspace_opts = WorkspaceOpts.extract_from_cli_options(other_opts)
    repository_opts = RepositoryOpts.extract_from_cli_options(other_opts)
    assert_no_remaining_opts(other_opts)
    return execute_stop_command(
        schedule_name=schedule_name,
        workspace_opts=workspace_opts,
        repository_opts=repository_opts,
        print_fn=click.echo,
    )


def execute_stop_command(
    *,
    schedule_name: str,
    workspace_opts: WorkspaceOpts,
    repository_opts: RepositoryOpts,
    print_fn: PrintFn,
    instance: Optional[DagsterInstance] = None,
):
    with (
        get_instance_for_cli() as instance,
        _get_repo(workspace_opts, repository_opts, instance) as repo,
    ):
        try:
            remote_schedule = repo.get_schedule(schedule_name)
            instance.stop_schedule(
                remote_schedule.get_remote_origin_id(),
                remote_schedule.selector_id,
                remote_schedule,
            )
        except DagsterInvariantViolationError as ex:
            raise click.UsageError(ex)  # pyright: ignore[reportArgumentType]

        print_fn(f"Stopped schedule {schedule_name}")


@schedule_cli.command(name="logs", help="Get logs for a schedule.")
@click.argument("schedule_name", required=False)
@workspace_options
@repository_options
def schedule_logs_command(schedule_name: Optional[str], **other_opts: object):
    workspace_opts = WorkspaceOpts.extract_from_cli_options(other_opts)
    repository_opts = RepositoryOpts.extract_from_cli_options(other_opts)
    assert_no_remaining_opts(other_opts)
    if schedule_name is None:
        print(  # noqa: T201
            "Noop: dagster schedule logs was called without any arguments specifying which "
            "schedules to retrieve logs for. Pass a schedule name"
        )
        return
    return execute_logs_command(
        schedule_name=schedule_name,
        workspace_opts=workspace_opts,
        repository_opts=repository_opts,
        print_fn=click.echo,
    )


def execute_logs_command(
    *,
    schedule_name: str,
    workspace_opts: WorkspaceOpts,
    repository_opts: RepositoryOpts,
    print_fn: PrintFn,
    instance: Optional[DagsterInstance] = None,
):
    with (
        get_instance_for_cli() as instance,
        _get_repo(workspace_opts, repository_opts, instance) as repo,
    ):
        if isinstance(instance.scheduler, DagsterDaemonScheduler):
            return print_fn(
                "This command is deprecated for the DagsterDaemonScheduler. "
                "Logs for the DagsterDaemonScheduler written to the process output. "
                "For help troubleshooting the Daemon Scheduler, see "
                "https://docs.dagster.io/guides/automate/schedules/troubleshooting-schedules"
            )

        logs_path = os.path.join(
            instance.logs_path_for_schedule(repo.get_schedule(schedule_name).get_remote_origin_id())
        )

        logs_directory = os.path.dirname(logs_path)
        result_files = glob.glob(f"{logs_directory}/*.result")
        most_recent_log = max(result_files, key=os.path.getctime) if result_files else None

        output = ""

        title = "Scheduler Logs:"
        output += "{title}\n{sep}\n{info}\n".format(
            title=title,
            sep="=" * len(title),
            info=logs_path,
        )

        title = (
            "Schedule Execution Logs:"
            "\nEvent logs from schedule execution. "
            "Errors that caused schedule executions to not run or fail can be found here. "
        )
        most_recent_info = (
            f"\nMost recent execution log: {most_recent_log}" if most_recent_log else ""
        )
        info = f"All execution logs: {logs_directory}{most_recent_info}"
        output += "\n{title}\n{sep}\n{info}\n".format(
            title=title,
            sep="=" * len(title),
            info=info,
        )

        print_fn(output)


@schedule_cli.command(name="restart", help="Restart a running schedule.")
@click.argument("schedule_name", required=False)
@click.option(
    "--restart-all-running",
    help="restart previously running schedules",
    is_flag=True,
    default=False,
)
@workspace_options
@repository_options
def schedule_restart_command(schedule_name: Optional[str], restart_all_running: bool, **other_opts):
    workspace_opts = WorkspaceOpts.extract_from_cli_options(other_opts)
    repository_opts = RepositoryOpts.extract_from_cli_options(other_opts)
    assert_no_remaining_opts(other_opts)
    return execute_restart_command(
        schedule_name=schedule_name,
        restart_all_running=restart_all_running,
        workspace_opts=workspace_opts,
        repository_opts=repository_opts,
        print_fn=click.echo,
    )


def execute_restart_command(
    *,
    schedule_name: Optional[str],
    restart_all_running: bool,
    workspace_opts: WorkspaceOpts,
    repository_opts: RepositoryOpts,
    print_fn: PrintFn,
):
    with (
        get_instance_for_cli() as instance,
        _get_repo(workspace_opts, repository_opts, instance) as repo,
    ):
        if restart_all_running:
            for schedule_state in instance.all_instigator_state(
                repo.get_remote_origin_id(),
                repo.selector_id,
                InstigatorType.SCHEDULE,
            ):
                if schedule_state.status == InstigatorStatus.RUNNING:
                    try:
                        remote_schedule = repo.get_schedule(schedule_state.instigator_name)
                        instance.stop_schedule(
                            schedule_state.instigator_origin_id,
                            remote_schedule.selector_id,
                            remote_schedule,
                        )
                        instance.start_schedule(remote_schedule)
                    except DagsterInvariantViolationError as ex:
                        raise click.UsageError(ex)  # pyright: ignore[reportArgumentType]

            print_fn(f"Restarted all running schedules for repository {repo.name}")
        else:
            if not schedule_name:
                raise click.UsageError("Missing schedule name")
            remote_schedule = repo.get_schedule(schedule_name)
            schedule_state = instance.get_instigator_state(
                remote_schedule.get_remote_origin_id(),
                remote_schedule.selector_id,
            )
            if schedule_state is not None and schedule_state.status != InstigatorStatus.RUNNING:
                click.UsageError(
                    f"Cannot restart a schedule {schedule_state.instigator_name} because is not currently running"
                )

            try:
                instance.stop_schedule(
                    schedule_state.instigator_origin_id,  # pyright: ignore[reportOptionalMemberAccess]
                    remote_schedule.selector_id,
                    remote_schedule,
                )
                instance.start_schedule(remote_schedule)
            except DagsterInvariantViolationError as ex:
                raise click.UsageError(ex)  # pyright: ignore[reportArgumentType]

            print_fn(f"Restarted schedule {schedule_name}")


@schedule_cli.command(name="wipe", help="Delete the schedule history and turn off all schedules.")
def schedule_wipe_command():
    return execute_wipe_command(click.echo)


def execute_wipe_command(print_fn: PrintFn):
    with get_instance_for_cli() as instance:
        confirmation = click.prompt(
            "Are you sure you want to turn off all schedules and delete all schedule history? Type"
            " DELETE"
        )
        if confirmation == "DELETE":
            instance.wipe_all_schedules()
            print_fn("Turned off all schedules and deleted all schedule history")
        else:
            print_fn("Exiting without turning off schedules or deleting schedule history")


@schedule_cli.command(name="debug", help="Debug information about the scheduler.")
def schedule_debug_command():
    return execute_debug_command(click.echo)


def execute_debug_command(print_fn: PrintFn):
    with get_instance_for_cli() as instance:
        debug_info = instance.scheduler_debug_info()

        output = ""

        errors = debug_info.errors
        if len(errors):
            title = "Errors (Run `dagster schedule up` to resolve)"
            output += "\n{title}\n{sep}\n{info}\n\n".format(
                title=title,
                sep="=" * len(title),
                info="\n".join(debug_info.errors),
            )

        title = "Scheduler Configuration"
        output += "{title}\n{sep}\n{info}\n".format(
            title=title,
            sep="=" * len(title),
            info=debug_info.scheduler_config_info,
        )

        title = "Scheduler Info"
        output += "{title}\n{sep}\n{info}\n".format(
            title=title, sep="=" * len(title), info=debug_info.scheduler_info
        )

        title = "Scheduler Storage Info"
        output += "\n{title}\n{sep}\n{info}\n".format(
            title=title,
            sep="=" * len(title),
            info="\n".join(debug_info.schedule_storage),
        )

        print_fn(output)


# ########################
# ##### HELPERS
# ########################


def print_changes(
    remote_repository: RemoteRepository,
    instance: DagsterInstance,
    print_fn: Callable[[object], None] = print,
    preview: bool = False,
) -> None:
    debug_info = instance.scheduler_debug_info()
    errors = debug_info.errors
    schedules = remote_repository.get_schedules()
    schedule_states = instance.all_instigator_state(
        remote_repository.get_remote_origin_id(),
        remote_repository.selector_id,
        InstigatorType.SCHEDULE,
    )
    schedules_dict = {s.get_remote_origin_id(): s for s in schedules}
    schedule_states_dict = {s.instigator_origin_id: s for s in schedule_states}

    schedule_origin_ids = set(schedules_dict.keys())
    schedule_state_ids = set(schedule_states_dict.keys())

    added_schedules = schedule_origin_ids - schedule_state_ids
    removed_schedules = schedule_state_ids - schedule_origin_ids

    changed_schedules = []
    for schedule_origin_id in schedule_origin_ids & schedule_state_ids:
        schedule_state = schedule_states_dict[schedule_origin_id]
        schedule = schedules_dict[schedule_origin_id]
        if schedule_state.instigator_data.cron_schedule != schedule.cron_schedule:  # type: ignore
            changed_schedules.append(schedule_origin_id)

    if not errors and not added_schedules and not changed_schedules and not removed_schedules:
        if preview:
            print_fn(click.style("No planned changes to schedules.", fg="magenta", bold=True))
            print_fn(f"{len(schedules)} schedules will remain unchanged")
        else:
            print_fn(click.style("No changes to schedules.", fg="magenta", bold=True))
            print_fn(f"{len(schedules)} schedules unchanged")
        return

    if errors:
        print_fn(
            click.style(
                "Planned Error Fixes:" if preview else "Errors Resolved:", fg="magenta", bold=True
            )
        )
        print_fn("\n".join(debug_info.errors))

    if added_schedules or changed_schedules or removed_schedules:
        print_fn(
            click.style(
                "Planned Schedule Changes:" if preview else "Changes:", fg="magenta", bold=True
            )
        )

    for schedule_origin_id in added_schedules:
        print_fn(
            click.style(
                f"  + {schedules_dict[schedule_origin_id].name} (add) [{schedule_origin_id}]",
                fg="green",
            )
        )

    for schedule_origin_id in changed_schedules:
        schedule_state = schedule_states_dict[schedule_origin_id]
        schedule = schedules_dict[schedule_origin_id]

        print_fn(
            click.style(
                f"  ~ {schedule.name} (update) [{schedule_origin_id}]",
                fg="yellow",
            )
        )
        print_fn(
            click.style("\t cron_schedule: ", fg="yellow")
            + click.style(schedule_state.instigator_data.cron_schedule, fg="red")  # type: ignore
            + " => "
            + click.style(schedule.cron_schedule, fg="green")
        )

    for schedule_origin_id in removed_schedules:
        print_fn(
            click.style(
                f"  - {schedule_states_dict[schedule_origin_id].instigator_name} (delete) [{schedule_origin_id}]",
                fg="red",
            )
        )


@contextmanager
def _get_repo(
    workspace_opts: WorkspaceOpts, repository_opts: RepositoryOpts, instance: DagsterInstance
) -> Iterator[RemoteRepository]:
    with get_repository_from_cli_opts(
        instance=instance,
        version=dagster_version,
        workspace_opts=workspace_opts,
        repository_opts=repository_opts,
    ) as repo:
        validate_repo_has_defined_schedules(repo)
        validate_dagster_home_is_set()
        yield repo



import datetime
from typing import Optional

from dagster._core.asset_graph_view.entity_subset import EntitySubset
from dagster._core.definitions.asset_key import AssetCheckKey, AssetKey
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationResult,
    BuiltinAutomationCondition,
)
from dagster._core.definitions.declarative_automation.automation_context import AutomationContext
from dagster._core.definitions.declarative_automation.operands.subset_automation_condition import (
    SubsetAutomationCondition,
)
from dagster._core.definitions.declarative_automation.utils import SerializableTimeDelta
from dagster._record import record
from dagster._serdes.serdes import whitelist_for_serdes
from dagster._utils.schedules import reverse_cron_string_iterator


@whitelist_for_serdes
@record
class CodeVersionChangedCondition(BuiltinAutomationCondition[AssetKey]):
    @property
    def name(self) -> str:
        return "code_version_changed"

    def evaluate(self, context: AutomationContext) -> AutomationResult[AssetKey]:
        previous_code_version = context.cursor
        current_code_version = context.asset_graph.get(context.key).code_version
        if previous_code_version is None or previous_code_version == current_code_version:
            true_subset = context.get_empty_subset()
        else:
            true_subset = context.candidate_subset

        return AutomationResult(context, true_subset, cursor=current_code_version)


@record
@whitelist_for_serdes
class InitialEvaluationCondition(BuiltinAutomationCondition):
    """Condition to determine if this is the initial evaluation of a given AutomationCondition."""

    @property
    def name(self) -> str:
        return "initial_evaluation"

    def evaluate(self, context: AutomationContext) -> AutomationResult:
        condition_tree_id = context.root_context.condition.get_unique_id()
        if context.previous_true_subset is None or condition_tree_id != context.cursor:
            subset = context.candidate_subset
        else:
            subset = context.get_empty_subset()
        return AutomationResult(context, subset, cursor=condition_tree_id)


@whitelist_for_serdes
@record
class MissingAutomationCondition(SubsetAutomationCondition):
    @property
    def name(self) -> str:
        return "missing"

    async def compute_subset(self, context: AutomationContext) -> EntitySubset:
        return await context.asset_graph_view.compute_missing_subset(
            key=context.key, from_subset=context.candidate_subset
        )


@whitelist_for_serdes(storage_name="InProgressAutomationCondition")
@record
class RunInProgressAutomationCondition(SubsetAutomationCondition):
    @property
    def name(self) -> str:
        return "run_in_progress"

    async def compute_subset(self, context: AutomationContext) -> EntitySubset:
        return await context.asset_graph_view.compute_run_in_progress_subset(key=context.key)


@whitelist_for_serdes
@record
class BackfillInProgressAutomationCondition(SubsetAutomationCondition):
    @property
    def name(self) -> str:
        return "backfill_in_progress"

    async def compute_subset(self, context: AutomationContext) -> EntitySubset:
        return await context.asset_graph_view.compute_backfill_in_progress_subset(key=context.key)


@whitelist_for_serdes(storage_name="FailedAutomationCondition")
@record
class ExecutionFailedAutomationCondition(SubsetAutomationCondition):
    @property
    def name(self) -> str:
        return "execution_failed"

    async def compute_subset(self, context: AutomationContext) -> EntitySubset:
        return await context.asset_graph_view.compute_execution_failed_subset(key=context.key)


@whitelist_for_serdes
@record
class WillBeRequestedCondition(SubsetAutomationCondition):
    @property
    def description(self) -> str:
        return "Will be requested this tick"

    @property
    def name(self) -> str:
        return "will_be_requested"

    def _executable_with_root_context_key(self, context: AutomationContext) -> bool:
        # TODO: once we can launch backfills via the asset daemon, this can be removed
        from dagster._core.definitions.asset_graph import executable_in_same_run

        root_key = context.root_context.key
        return executable_in_same_run(
            asset_graph=context.asset_graph_view.asset_graph,
            child_key=root_key,
            parent_key=context.key,
        )

    def compute_subset(self, context: AutomationContext) -> EntitySubset:
        current_result = context.request_subsets_by_key.get(context.key)
        if current_result and self._executable_with_root_context_key(context):
            return current_result
        else:
            return context.get_empty_subset()


@whitelist_for_serdes
@record
class NewlyRequestedCondition(SubsetAutomationCondition):
    @property
    def name(self) -> str:
        return "newly_requested"

    def compute_subset(self, context: AutomationContext) -> EntitySubset:
        return context.previous_requested_subset or context.get_empty_subset()


@whitelist_for_serdes
@record
class LatestRunExecutedWithRootTargetCondition(SubsetAutomationCondition):
    @property
    def name(self) -> str:
        return "executed_with_root_target"

    async def compute_subset(self, context: AutomationContext) -> EntitySubset:
        return await context.asset_graph_view.compute_latest_run_executed_with_subset(
            from_subset=context.candidate_subset, target=context.root_context.key
        )


@whitelist_for_serdes
@record
class NewlyUpdatedCondition(SubsetAutomationCondition):
    @property
    def name(self) -> str:
        return "newly_updated"

    async def compute_subset(self, context: AutomationContext) -> EntitySubset:
        # if it's the first time evaluating, just return the empty subset
        if context.previous_temporal_context is None:
            return context.get_empty_subset()
        return await context.asset_graph_view.compute_updated_since_temporal_context_subset(
            key=context.key, temporal_context=context.previous_temporal_context
        )


@whitelist_for_serdes
@record
class DataVersionChangedCondition(SubsetAutomationCondition):
    @property
    def name(self) -> str:
        return "data_version_changed"

    async def compute_subset(self, context: AutomationContext) -> EntitySubset:
        # if it's the first time evaluating, just return the empty subset
        if context.previous_temporal_context is None:
            return context.get_empty_subset()
        return await context.asset_graph_view.compute_data_version_changed_since_temporal_context_subset(
            key=context.key, temporal_context=context.previous_temporal_context
        )


@whitelist_for_serdes
@record
class CronTickPassedCondition(SubsetAutomationCondition):
    cron_schedule: str
    cron_timezone: str

    @property
    def name(self) -> str:
        return f"cron_tick_passed(cron_schedule={self.cron_schedule}, cron_timezone={self.cron_timezone})"

    def _get_previous_cron_tick(self, effective_dt: datetime.datetime) -> datetime.datetime:
        previous_ticks = reverse_cron_string_iterator(
            end_timestamp=effective_dt.timestamp(),
            cron_string=self.cron_schedule,
            execution_timezone=self.cron_timezone,
        )
        return next(previous_ticks)

    def compute_subset(self, context: AutomationContext) -> EntitySubset:
        previous_cron_tick = self._get_previous_cron_tick(context.evaluation_time)
        if (
            # no previous evaluation
            context.previous_evaluation_time is None
            # cron tick was not newly passed
            or previous_cron_tick < context.previous_evaluation_time
        ):
            return context.get_empty_subset()
        else:
            return context.candidate_subset


@whitelist_for_serdes
@record
class InLatestTimeWindowCondition(SubsetAutomationCondition):
    serializable_lookback_timedelta: Optional[SerializableTimeDelta] = None

    @staticmethod
    def from_lookback_delta(
        lookback_delta: Optional[datetime.timedelta],
    ) -> "InLatestTimeWindowCondition":
        return InLatestTimeWindowCondition(
            serializable_lookback_timedelta=SerializableTimeDelta.from_timedelta(lookback_delta)
            if lookback_delta
            else None
        )

    @property
    def lookback_timedelta(self) -> Optional[datetime.timedelta]:
        return (
            self.serializable_lookback_timedelta.to_timedelta()
            if self.serializable_lookback_timedelta
            else None
        )

    @property
    def description(self) -> str:
        return (
            f"Within {self.lookback_timedelta} of the end of the latest time window"
            if self.lookback_timedelta
            else "Within latest time window"
        )

    @property
    def name(self) -> str:
        name = "in_latest_time_window"
        if self.serializable_lookback_timedelta:
            name += f"(lookback_timedelta={self.lookback_timedelta})"
        return name

    def compute_subset(self, context: AutomationContext) -> EntitySubset:
        return context.asset_graph_view.compute_latest_time_window_subset(
            context.key, lookback_delta=self.lookback_timedelta
        )


@whitelist_for_serdes
@record
class CheckResultCondition(SubsetAutomationCondition[AssetCheckKey]):
    passed: bool

    @property
    def name(self) -> str:
        return "check_passed" if self.passed else "check_failed"

    async def compute_subset(
        self, context: AutomationContext[AssetCheckKey]
    ) -> EntitySubset[AssetCheckKey]:
        from dagster._core.storage.asset_check_execution_record import (
            AssetCheckExecutionResolvedStatus,
        )

        target_status = (
            AssetCheckExecutionResolvedStatus.SUCCEEDED
            if self.passed
            else AssetCheckExecutionResolvedStatus.FAILED
        )
        return await context.asset_graph_view.compute_subset_with_status(
            key=context.key, status=target_status
        )



import os
import sys
from collections.abc import Iterator
from contextlib import contextmanager
from typing import Optional

import click

from dagster import (
    DagsterInvariantViolationError,
    __version__ as dagster_version,
)
from dagster._cli.utils import (
    assert_no_remaining_opts,
    get_instance_for_cli,
    validate_dagster_home_is_set,
    validate_repo_has_defined_sensors,
)
from dagster._cli.workspace.cli_target import (
    RepositoryOpts,
    WorkspaceOpts,
    get_code_location_from_workspace,
    get_remote_repository_from_code_location,
    get_repository_from_cli_opts,
    get_workspace_from_cli_opts,
    repository_options,
    workspace_options,
)
from dagster._core.definitions.run_request import InstigatorType
from dagster._core.instance import DagsterInstance
from dagster._core.remote_representation import RemoteRepository
from dagster._core.scheduler.instigation import (
    InstigatorState,
    InstigatorStatus,
    SensorInstigatorData,
)
from dagster._utils import PrintFn
from dagster._utils.error import serializable_error_info_from_exc_info
from dagster._utils.yaml_utils import dump_run_config_yaml


@click.group(name="sensor")
def sensor_cli():
    """Commands for working with Dagster sensors."""


@sensor_cli.command(
    name="list",
    help="List all sensors that correspond to a repository.",
)
@click.option(
    "--running", "running_filter", help="Filter for running sensors", is_flag=True, default=False
)
@click.option(
    "--stopped", "stopped_filter", help="Filter for stopped sensors", is_flag=True, default=False
)
@click.option(
    "--name", "name_filter", help="Only display sensor sensor names", is_flag=True, default=False
)
@workspace_options
@repository_options
def sensor_list_command(
    running_filter: bool, stopped_filter: bool, name_filter: bool, **other_opts: object
):
    workspace_opts = WorkspaceOpts.extract_from_cli_options(other_opts)
    repository_opts = RepositoryOpts.extract_from_cli_options(other_opts)
    assert_no_remaining_opts(other_opts)
    return execute_list_command(
        running_filter=running_filter,
        stopped_filter=stopped_filter,
        name_filter=name_filter,
        workspace_opts=workspace_opts,
        repository_opts=repository_opts,
        print_fn=click.echo,
    )


def execute_list_command(
    *,
    running_filter: bool,
    stopped_filter: bool,
    name_filter: bool,
    workspace_opts: WorkspaceOpts,
    repository_opts: RepositoryOpts,
    print_fn: PrintFn,
):
    with (
        get_instance_for_cli() as instance,
        _get_repo(workspace_opts, repository_opts, instance) as repo,
    ):
        if not name_filter:
            title = f"Repository {repo.name}"
            print_fn(title)
            print_fn("*" * len(title))

        repo_sensors = repo.get_sensors()
        stored_sensors_by_origin_id = {
            stored_sensor_state.instigator_origin_id: stored_sensor_state
            for stored_sensor_state in instance.all_instigator_state(
                repo.get_remote_origin_id(),
                repo.selector_id,
                instigator_type=InstigatorType.SENSOR,
            )
        }

        first = True

        for sensor in repo_sensors:
            sensor_state = sensor.get_current_instigator_state(
                stored_sensors_by_origin_id.get(sensor.get_remote_origin_id())
            )

            if running_filter and not sensor_state.is_running:
                continue
            if stopped_filter and sensor_state.is_running:
                continue

            if name_filter:
                print_fn(sensor.name)
                continue

            status = "RUNNING" if sensor_state.is_running else "STOPPED"
            sensor_title = f"Sensor: {sensor.name} [{status}]"
            if not first:
                print_fn("*" * len(sensor_title))

            first = False

            print_fn(sensor_title)


@sensor_cli.command(name="start", help="Start an existing sensor.")
@click.argument("sensor_name", required=False)
@click.option("--start-all", help="start all sensors", is_flag=True, default=False)
@workspace_options
@repository_options
def sensor_start_command(sensor_name: Optional[str], start_all: bool, **other_opts: object):
    workspace_opts = WorkspaceOpts.extract_from_cli_options(other_opts)
    repository_opts = RepositoryOpts.extract_from_cli_options(other_opts)
    assert_no_remaining_opts(other_opts)
    return execute_start_command(
        sensor_name=sensor_name,
        start_all=start_all,
        workspace_opts=workspace_opts,
        repository_opts=repository_opts,
        print_fn=click.echo,
    )


def execute_start_command(
    *,
    sensor_name: Optional[str],
    start_all: bool,
    workspace_opts: WorkspaceOpts,
    repository_opts: RepositoryOpts,
    print_fn: PrintFn,
):
    with (
        get_instance_for_cli() as instance,
        _get_repo(workspace_opts, repository_opts, instance) as repo,
    ):
        if start_all:
            try:
                for sensor in repo.get_sensors():
                    instance.start_sensor(sensor)
                print_fn(f"Started all sensors for repository {repo.name}")
            except DagsterInvariantViolationError as ex:
                raise click.UsageError(ex)  # pyright: ignore[reportArgumentType]
        else:
            if not sensor_name:
                raise click.UsageError("Missing sensor name argument")
            try:
                sensor = repo.get_sensor(sensor_name)
                instance.start_sensor(sensor)
            except DagsterInvariantViolationError as ex:
                raise click.UsageError(ex)  # pyright: ignore[reportArgumentType]

            print_fn(f"Started sensor {sensor_name}")


@sensor_cli.command(name="stop", help="Stop an existing sensor.")
@click.argument("sensor_name")
@workspace_options
@repository_options
def sensor_stop_command(sensor_name: str, **other_opts: object):
    workspace_opts = WorkspaceOpts.extract_from_cli_options(other_opts)
    repository_opts = RepositoryOpts.extract_from_cli_options(other_opts)
    assert_no_remaining_opts(other_opts)
    return execute_stop_command(
        sensor_name=sensor_name,
        workspace_opts=workspace_opts,
        repository_opts=repository_opts,
        print_fn=click.echo,
    )


def execute_stop_command(
    *,
    sensor_name: str,
    workspace_opts: WorkspaceOpts,
    repository_opts: RepositoryOpts,
    print_fn: PrintFn,
):
    with (
        get_instance_for_cli() as instance,
        _get_repo(workspace_opts, repository_opts, instance) as repo,
    ):
        try:
            sensor = repo.get_sensor(sensor_name)
            instance.stop_sensor(
                sensor.get_remote_origin_id(),
                sensor.selector_id,
                sensor,
            )
        except DagsterInvariantViolationError as ex:
            raise click.UsageError(ex)  # pyright: ignore[reportArgumentType]

        print_fn(f"Stopped sensor {sensor_name}")


@sensor_cli.command(name="preview", help="Preview an existing sensor execution.")
@click.argument("sensor_name")
@click.option(
    "--since",
    type=float,
    help="Set the last_tick_completion_time value as a timestamp float for the sensor context",
    default=None,
)
@click.option(
    "--last_run_key",
    help="Set the last_run_key value for the sensor context",
    default=None,
)
@click.option(
    "--cursor",
    help="Set the cursor value for the sensor context",
    default=None,
)
@workspace_options
@repository_options
def sensor_preview_command(
    sensor_name: str,
    since: Optional[float],
    last_run_key: Optional[str],
    cursor: Optional[str],
    **other_opts: object,
):
    workspace_opts = WorkspaceOpts.extract_from_cli_options(other_opts)
    repository_opts = RepositoryOpts.extract_from_cli_options(other_opts)
    assert_no_remaining_opts(other_opts)

    return execute_preview_command(
        sensor_name=sensor_name,
        since=since,
        last_run_key=last_run_key,
        cursor=cursor,
        workspace_opts=workspace_opts,
        repository_opts=repository_opts,
        print_fn=click.echo,
    )


def execute_preview_command(
    *,
    sensor_name: str,
    since: Optional[float],
    last_run_key: Optional[str],
    cursor: Optional[str],
    workspace_opts: WorkspaceOpts,
    repository_opts: RepositoryOpts,
    print_fn: PrintFn,
    instance: Optional[DagsterInstance] = None,
):
    # We don't call _get_repo here because we need the code location.
    with (
        get_instance_for_cli() as instance,
        get_workspace_from_cli_opts(
            instance, version=dagster_version, workspace_opts=workspace_opts
        ) as workspace,
    ):
        code_location = get_code_location_from_workspace(workspace, repository_opts.location)
        repo = get_remote_repository_from_code_location(code_location, repository_opts.repository)
        validate_repo_has_defined_sensors(repo)
        validate_dagster_home_is_set()
        try:
            sensor = repo.get_sensor(sensor_name)
            sensor_runtime_data = code_location.get_sensor_execution_data(
                instance,
                repo.handle,
                sensor.name,
                since,
                last_run_key,
                cursor,
                log_key=None,
                last_sensor_start_time=None,
            )
        except Exception:
            error_info = serializable_error_info_from_exc_info(sys.exc_info())
            print_fn(f"Failed to resolve sensor for {sensor.name} : {error_info.to_string()}")
            return

        if not sensor_runtime_data.run_requests:
            if sensor_runtime_data.skip_message:
                print_fn(
                    f"Sensor returned false for {sensor.name}, skipping: {sensor_runtime_data.skip_message}"
                )
            else:
                print_fn(f"Sensor returned false for {sensor.name}, skipping")
        else:
            print_fn(
                "Sensor returning run requests for {num} run(s):\n\n{run_requests}".format(
                    num=len(sensor_runtime_data.run_requests),
                    run_requests="\n".join(
                        dump_run_config_yaml(run_request.run_config)
                        for run_request in sensor_runtime_data.run_requests
                    ),
                )
            )


@sensor_cli.command(name="cursor", help="Set the cursor value for an existing sensor.")
@click.argument("sensor_name")
@click.option(
    "--set",
    "cursor_value",
    help="Set the cursor value for the sensor context",
    default=None,
)
@click.option(
    "--delete", help="Delete the existing cursor value for the sensor context", is_flag=True
)
@workspace_options
@repository_options
def sensor_cursor_command(
    sensor_name: str, cursor_value: Optional[str], delete: bool, **other_opts: object
):
    workspace_opts = WorkspaceOpts.extract_from_cli_options(other_opts)
    repository_opts = RepositoryOpts.extract_from_cli_options(other_opts)
    assert_no_remaining_opts(other_opts)
    return execute_cursor_command(
        sensor_name=sensor_name,
        cursor_value=cursor_value,
        delete=delete,
        workspace_opts=workspace_opts,
        repository_opts=repository_opts,
        print_fn=click.echo,
    )


def execute_cursor_command(
    *,
    sensor_name: str,
    cursor_value: Optional[str],
    delete: bool,
    workspace_opts: WorkspaceOpts,
    repository_opts: RepositoryOpts,
    print_fn: PrintFn,
):
    if bool(delete) == bool(cursor_value):
        # must use one of delete/set
        raise click.UsageError("Must set cursor using `--set <value>` or use `--delete`")

    with (
        get_instance_for_cli() as instance,
        _get_repo(workspace_opts, repository_opts, instance) as repo,
    ):
        sensor = repo.get_sensor(sensor_name)
        job_state = instance.get_instigator_state(sensor.get_remote_origin_id(), sensor.selector_id)
        if not job_state:
            instance.add_instigator_state(
                InstigatorState(
                    sensor.get_remote_origin(),
                    InstigatorType.SENSOR,
                    InstigatorStatus.STOPPED,
                    SensorInstigatorData(
                        min_interval=sensor.min_interval_seconds,
                        cursor=cursor_value,
                        sensor_type=sensor.sensor_type,
                    ),
                )
            )
        else:
            instance.update_instigator_state(
                job_state.with_data(
                    SensorInstigatorData(
                        last_tick_timestamp=job_state.instigator_data.last_tick_timestamp,  # pyright: ignore[reportOptionalMemberAccess,reportAttributeAccessIssue]
                        last_run_key=job_state.instigator_data.last_run_key,  # pyright: ignore[reportOptionalMemberAccess,reportAttributeAccessIssue]
                        min_interval=sensor.min_interval_seconds,
                        cursor=cursor_value,
                        last_tick_start_timestamp=job_state.instigator_data.last_tick_start_timestamp,  # pyright: ignore[reportOptionalMemberAccess,reportAttributeAccessIssue]
                        last_sensor_start_timestamp=job_state.instigator_data.last_sensor_start_timestamp,  # pyright: ignore[reportOptionalMemberAccess,reportAttributeAccessIssue]
                        sensor_type=sensor.sensor_type,
                    ),
                )
            )
        if cursor_value:
            print_fn(f'Set cursor state for sensor {sensor.name} to "{cursor_value}"')
        else:
            print_fn(f"Cleared cursor state for sensor {sensor.name}")


# ########################
# ##### HELPERS
# ########################


def print_changes(
    remote_repository: RemoteRepository,
    instance: DagsterInstance,
    print_fn: PrintFn = print,
    preview: bool = False,
) -> None:
    sensor_states = instance.all_instigator_state(
        remote_repository.get_remote_origin_id(),
        remote_repository.selector_id,
        InstigatorType.SENSOR,
    )
    sensors = remote_repository.get_sensors()
    sensors_dict = {s.get_remote_origin_id(): s for s in sensors}
    sensor_states_dict = {s.instigator_origin_id: s for s in sensor_states}

    sensor_origin_ids = set(sensors_dict.keys())
    sensor_state_ids = set(sensor_states_dict.keys())

    added_sensors = sensor_origin_ids - sensor_state_ids
    removed_sensors = sensor_state_ids - sensor_origin_ids

    if not added_sensors and not removed_sensors:
        if preview:
            print_fn(click.style("No planned changes to sensors.", fg="magenta", bold=True))
            print_fn(f"{len(sensors)} sensors will remain unchanged")
        else:
            print_fn(click.style("No changes to sensors.", fg="magenta", bold=True))
            print_fn(f"{len(sensors)} sensors unchanged")
        return

    print_fn(
        click.style("Planned Sensor Changes:" if preview else "Changes:", fg="magenta", bold=True)
    )

    for sensor_origin_id in added_sensors:
        print_fn(
            click.style(
                f"  + {sensors_dict[sensor_origin_id].name} (add) [{sensor_origin_id}]",
                fg="green",
            )
        )

    for sensor_origin_id in removed_sensors:
        print_fn(
            click.style(
                f"  + {sensors_dict[sensor_origin_id].name} (delete) [{sensor_origin_id}]",
                fg="red",
            )
        )


@contextmanager
def _get_repo(
    workspace_opts: WorkspaceOpts, repository_opts: RepositoryOpts, instance: DagsterInstance
) -> Iterator[RemoteRepository]:
    with get_repository_from_cli_opts(
        instance=instance,
        version=dagster_version,
        workspace_opts=workspace_opts,
        repository_opts=repository_opts,
    ) as repo:
        validate_repo_has_defined_sensors(repo)
        validate_dagster_home_is_set()
        yield repo


def check_repo_and_scheduler(repository: RemoteRepository, instance: DagsterInstance) -> None:
    repository_name = repository.name

    if not repository.get_sensors():
        raise click.UsageError(f"There are no sensors defined for repository {repository_name}.")

    if not os.getenv("DAGSTER_HOME"):
        raise click.UsageError(
            "The environment variable $DAGSTER_HOME is not set. Dagster requires this "
            "environment variable to be set to an existing directory in your filesystem "
            "that contains your dagster instance configuration file (dagster.yaml).\n"
            "You can resolve this error by exporting the environment variable."
            "For example, you can run the following command in your shell or "
            "include it in your shell configuration file:\n"
            '\texport DAGSTER_HOME="~/dagster_home"'
            "\n\n"
        )



from collections.abc import Mapping
from typing import Any, Optional, Union

from dagster._core.definitions.asset_key import AssetKey
from dagster._record import record
from dagster._utils.merger import merge_dicts


@record
class OutputMetadataHandle:
    output_name: str
    mapping_key: Optional[str]


@record
class AssetMetadataHandle:
    asset_key: AssetKey
    partition_key: Optional[str]


@record
class OutputMetadataAccumulator:
    per_output_metadata: Mapping[
        Union[OutputMetadataHandle, AssetMetadataHandle], Mapping[str, Any]
    ]

    @staticmethod
    def empty() -> "OutputMetadataAccumulator":
        return OutputMetadataAccumulator(per_output_metadata={})

    def get_output_metadata(
        self, output_name: str, mapping_key: Optional[str]
    ) -> Mapping[str, Any]:
        handle = OutputMetadataHandle(
            output_name=output_name,
            mapping_key=mapping_key,
        )
        return self.per_output_metadata.get(handle, {})

    def get_asset_metadata(
        self, asset_key: AssetKey, partition_key: Optional[str]
    ) -> Mapping[str, Any]:
        handle = AssetMetadataHandle(
            asset_key=asset_key,
            partition_key=partition_key,
        )
        return self.per_output_metadata.get(handle, {})

    def with_additional_output_metadata(
        self,
        output_name: str,
        mapping_key: Optional[str],
        metadata: Mapping[str, Any],
    ) -> "OutputMetadataAccumulator":
        return self._with_metadata(
            handle=OutputMetadataHandle(
                output_name=output_name,
                mapping_key=mapping_key,
            ),
            metadata=metadata,
        )

    def _with_metadata(
        self, handle: Union[OutputMetadataHandle, AssetMetadataHandle], metadata: Mapping[str, Any]
    ) -> "OutputMetadataAccumulator":
        return OutputMetadataAccumulator(
            per_output_metadata=merge_dicts(
                self.per_output_metadata,
                {handle: merge_dicts(self.per_output_metadata.get(handle, {}), metadata)},
            )
        )

    def with_additional_asset_metadata(
        self,
        asset_key: AssetKey,
        partition_key: Optional[str],
        metadata: Mapping[str, Any],
    ) -> "OutputMetadataAccumulator":
        return self._with_metadata(
            handle=AssetMetadataHandle(
                asset_key=asset_key,
                partition_key=partition_key,
            ),
            metadata=metadata,
        )



from dagster._generate.download import download_example_from_github as download_example_from_github
from dagster._generate.generate import (
    generate_project as generate_project,
    generate_repository as generate_repository,
)



import logging
import os
import sys
from collections.abc import Iterator, Mapping, Sequence
from contextlib import contextmanager
from typing import Any, Callable, Optional, TypeVar, cast

import click
from click import UsageError
from typing_extensions import Never, Self, TypeAlias

import dagster._check as check
from dagster import __version__ as dagster_version
from dagster._cli.utils import has_pyproject_dagster_block, serialize_sorted_quoted
from dagster._core.code_pointer import CodePointer
from dagster._core.definitions.reconstruct import repository_def_from_target_def
from dagster._core.instance import DagsterInstance
from dagster._core.origin import DEFAULT_DAGSTER_ENTRY_POINT, RepositoryPythonOrigin
from dagster._core.remote_representation.code_location import CodeLocation
from dagster._core.remote_representation.external import RemoteJob, RemoteRepository
from dagster._core.remote_representation.origin import (
    CodeLocationOrigin,
    GrpcServerCodeLocationOrigin,
    InProcessCodeLocationOrigin,
)
from dagster._core.workspace.context import WorkspaceProcessContext, WorkspaceRequestContext
from dagster._core.workspace.load_target import (
    CompositeTarget,
    EmptyWorkspaceTarget,
    GrpcServerTarget,
    InProcessWorkspaceLoadTarget,
    ModuleTarget,
    PackageTarget,
    PyProjectFileTarget,
    PythonFileTarget,
    WorkspaceFileTarget,
    WorkspaceLoadTarget,
)
from dagster._grpc.utils import get_loadable_targets
from dagster._record import record
from dagster._seven import JSONDecodeError, json
from dagster._utils.error import serializable_error_info_from_exc_info
from dagster._utils.yaml_utils import load_yaml_from_glob_list

logger = logging.getLogger("dagster")
WORKSPACE_TARGET_WARNING = (
    "Can only use ONE of --workspace/-w, --python-file/-f, --module-name/-m, --grpc-port,"
    " --grpc-socket."
)

T = TypeVar("T")
T_Callable = TypeVar("T_Callable", bound=Callable[..., Any])


WORKSPACE_CLI_ARGS = (
    "workspace",
    "python_file",
    "working_directory",
    "package_name",
    "module_name",
    "attribute",
    "grpc_host",
    "grpc_port",
    "grpc_socket",
)


def _get_workspace_load_target_from_cli_opts(
    workspace_opts: "WorkspaceOpts",
) -> WorkspaceLoadTarget:
    if _are_attrs_falsey(workspace_opts, *WORKSPACE_CLI_ARGS):
        if workspace_opts.empty_workspace:
            return EmptyWorkspaceTarget()
        elif has_pyproject_dagster_block("pyproject.toml"):
            return PyProjectFileTarget("pyproject.toml")
        elif os.path.exists("workspace.yaml"):
            return WorkspaceFileTarget(paths=["workspace.yaml"])
        else:
            raise click.UsageError(
                "No arguments given and no [tool.dagster] block in pyproject.toml found."
            )

    if workspace_opts.workspace:
        _check_attrs_falsey(
            workspace_opts,
            *(k for k in WORKSPACE_CLI_ARGS if k not in ["workspace"]),
        )
        return WorkspaceFileTarget(paths=list(workspace_opts.workspace))

    elif workspace_opts.python_file:
        _check_attrs_falsey(
            workspace_opts,
            *(
                k
                for k in WORKSPACE_CLI_ARGS
                if k not in ["python_file", "attribute", "working_directory"]
            ),
        )
        working_directory = workspace_opts.working_directory or os.getcwd()

        if len(workspace_opts.python_file) == 1:
            return PythonFileTarget(
                python_file=workspace_opts.python_file[0],
                attribute=workspace_opts.attribute,
                working_directory=working_directory,
                location_name=None,
            )
        else:
            # multiple files

            if workspace_opts.attribute:
                raise UsageError(
                    "If you are specifying multiple files you cannot specify an attribute."
                )

            return CompositeTarget(
                targets=[
                    PythonFileTarget(
                        python_file=python_file,
                        attribute=None,
                        working_directory=working_directory,
                        location_name=None,
                    )
                    for python_file in workspace_opts.python_file
                ]
            )

    elif workspace_opts.module_name:
        _check_attrs_falsey(
            workspace_opts,
            *(
                k
                for k in WORKSPACE_CLI_ARGS
                if k not in ["module_name", "attribute", "working_directory"]
            ),
        )

        working_directory = workspace_opts.working_directory or os.getcwd()

        if len(workspace_opts.module_name) == 1:
            return ModuleTarget(
                module_name=workspace_opts.module_name[0],
                attribute=workspace_opts.attribute,
                working_directory=working_directory,
                location_name=None,
            )
        else:
            # multiple modules

            if workspace_opts.attribute:
                raise UsageError(
                    "If you are specifying multiple modules you cannot specify an attribute. Got"
                    f" modules {workspace_opts.module_name}."
                )

            return CompositeTarget(
                targets=[
                    ModuleTarget(
                        module_name=module_name,
                        attribute=None,
                        working_directory=working_directory,
                        location_name=None,
                    )
                    for module_name in workspace_opts.module_name
                ]
            )

    elif workspace_opts.package_name:
        _check_attrs_falsey(
            workspace_opts,
            *(
                k
                for k in WORKSPACE_CLI_ARGS
                if k not in ["package_name", "attribute", "working_directory"]
            ),
        )
        working_directory = workspace_opts.working_directory or os.getcwd()

        if len(workspace_opts.package_name) == 1:
            return PackageTarget(
                package_name=workspace_opts.package_name[0],
                attribute=workspace_opts.attribute,
                working_directory=working_directory,
                location_name=None,
            )
        else:
            if workspace_opts.attribute:
                raise UsageError(
                    "If you are specifying multiple packages you cannot specify an attribute. Got"
                    f" packages {workspace_opts.package_name}."
                )

            return CompositeTarget(
                targets=[
                    PackageTarget(
                        package_name=package_name,
                        attribute=None,
                        working_directory=working_directory,
                        location_name=None,
                    )
                    for package_name in workspace_opts.package_name
                ]
            )

    elif workspace_opts.grpc_port:
        _check_attrs_falsey(
            workspace_opts,
            *(k for k in WORKSPACE_CLI_ARGS if k not in ["grpc_port", "grpc_host"]),
        )
        return GrpcServerTarget(
            port=workspace_opts.grpc_port,
            socket=None,
            host=workspace_opts.grpc_host or "localhost",
            location_name=None,
        )
    elif workspace_opts.grpc_socket:
        _check_attrs_falsey(
            workspace_opts,
            *(k for k in WORKSPACE_CLI_ARGS if k not in ["grpc_socket", "grpc_host"]),
        )
        return GrpcServerTarget(
            port=None,
            socket=workspace_opts.grpc_socket,
            host=workspace_opts.grpc_host or "localhost",
            location_name=None,
        )
    else:
        _raise_cli_usage_error()


@contextmanager
def get_workspace_from_cli_opts(
    instance: DagsterInstance,
    version: str,
    workspace_opts: "WorkspaceOpts",
    allow_in_process: bool = False,
) -> Iterator[WorkspaceRequestContext]:
    load_target = workspace_opts.to_load_target(allow_in_process)
    if isinstance(load_target, InProcessWorkspaceLoadTarget):
        logger.debug("Loading workspace in-process")
    else:
        logger.debug("Loading workspace with gRPC server")

    with WorkspaceProcessContext(
        instance=instance,
        version=version,
        read_only=False,
        workspace_load_target=load_target,
    ) as workspace_process_context:
        yield workspace_process_context.create_request_context()


@contextmanager
def get_repository_from_cli_opts(
    instance: DagsterInstance,
    version: str,
    workspace_opts: "WorkspaceOpts",
    repository_opts: Optional["RepositoryOpts"],
) -> Iterator[RemoteRepository]:
    # Instance isn't strictly required to load a RemoteRepository, but is included
    # to satisfy the WorkspaceProcessContext / WorkspaceRequestContext requirements
    with get_workspace_from_cli_opts(
        instance, version=dagster_version, workspace_opts=workspace_opts
    ) as workspace:
        repository_name = repository_opts.repository if repository_opts else None
        location_name = repository_opts.location if repository_opts else None
        code_location = get_code_location_from_workspace(workspace, location_name)
        yield get_remote_repository_from_code_location(code_location, repository_name)


@contextmanager
def get_job_from_cli_opts(
    instance: DagsterInstance,
    version: str,
    workspace_opts: "WorkspaceOpts",
    repository_opts: Optional["RepositoryOpts"],
    job_name: Optional[str],
) -> Iterator[RemoteJob]:
    # Instance isn't strictly required to load an RemoteJob, but is included
    # to satisfy the WorkspaceProcessContext / WorkspaceRequestContext requirements
    with get_repository_from_cli_opts(instance, version, workspace_opts, repository_opts) as repo:
        yield get_remote_job_from_remote_repo(repo, job_name)


# ########################
# ##### VALUE OBJECTS
# ########################

# These classes correspond to the reusable option groups defined in the decorators below. When one
# of these decorators is used, the resulting options should be immediately parsed into the
# corresponding value object at the top of the click command body, by calling
# `extract_from_cli_options`.


@record
class PythonPointerOpts:
    python_file: Optional[str] = None
    module_name: Optional[str] = None
    package_name: Optional[str] = None
    working_directory: Optional[str] = None
    attribute: Optional[str] = None

    @classmethod
    def extract_from_cli_options(cls, cli_options: dict[str, Any]) -> Self:
        # This is expected to always be called from a click entry point, so all options should be
        # present in the dictionary. We rely on `@record` for type-checking.
        return cls(
            python_file=cli_options.pop("python_file"),
            module_name=cli_options.pop("module_name"),
            package_name=cli_options.pop("package_name"),
            working_directory=cli_options.pop("working_directory"),
            attribute=cli_options.pop("attribute"),
        )

    def to_workspace_opts(self) -> "WorkspaceOpts":
        return WorkspaceOpts(
            python_file=(self.python_file,) if self.python_file else None,
            module_name=(self.module_name,) if self.module_name else None,
            package_name=(self.package_name,) if self.package_name else None,
            working_directory=self.working_directory,
            attribute=self.attribute,
        )


@record
class WorkspaceOpts:
    empty_workspace: bool = False
    workspace: Optional[Sequence[str]] = None

    # Like PythonPointerParams but multiple files/modules/packages are allowed
    python_file: Optional[Sequence[str]] = None
    module_name: Optional[Sequence[str]] = None
    package_name: Optional[Sequence[str]] = None
    working_directory: Optional[str] = None
    attribute: Optional[str] = None

    # For gRPC server
    grpc_port: Optional[int] = None
    grpc_socket: Optional[str] = None
    grpc_host: Optional[str] = None
    use_ssl: bool = False

    @classmethod
    def extract_from_cli_options(cls, cli_options: dict[str, Any]) -> Self:
        # This is expected to always be called from a click entry point, so all options should be
        # present in the dictionary. We rely on `@record` for type-checking.
        return cls(
            empty_workspace=cli_options.pop("empty_workspace"),
            workspace=cli_options.pop("workspace"),
            python_file=cli_options.pop("python_file"),
            module_name=cli_options.pop("module_name"),
            package_name=cli_options.pop("package_name"),
            working_directory=cli_options.pop("working_directory"),
            attribute=cli_options.pop("attribute"),
            grpc_port=cli_options.pop("grpc_port"),
            grpc_socket=cli_options.pop("grpc_socket"),
            grpc_host=cli_options.pop("grpc_host"),
            use_ssl=cli_options.pop("use_ssl"),
        )

    def to_load_target(self, allow_in_process: bool = False) -> WorkspaceLoadTarget:
        load_target = _get_workspace_load_target_from_cli_opts(self)
        origins = load_target.create_origins()

        # We can load a workspace in-process if there is only one origin and the python executable is
        # unspecified or matches that of the current process.
        origins = load_target.create_origins()
        can_load_in_process = len(origins) == 1 and _origin_executable_matches_current_process(
            origins[0]
        )
        if allow_in_process and can_load_in_process:
            origin = origins[0]
            return InProcessWorkspaceLoadTarget(
                [
                    InProcessCodeLocationOrigin(
                        origin.loadable_target_origin,
                        container_image=None,
                        location_name=origin.location_name,
                    )
                ]
            )
        else:
            return load_target


def _origin_executable_matches_current_process(origin: CodeLocationOrigin) -> bool:
    # loadable_target_origin is unknown for GrpcServerCodeLocationOrigin
    return not isinstance(origin, GrpcServerCodeLocationOrigin) and (
        origin.loadable_target_origin.executable_path is None
        or origin.loadable_target_origin.executable_path == sys.executable
    )


@record
class RepositoryOpts:
    repository: Optional[str] = None
    location: Optional[str] = None

    @classmethod
    def extract_from_cli_options(cls, cli_options: dict[str, object]) -> Self:
        return cls(
            repository=check.opt_inst(cli_options.pop("repository"), str),
            location=check.opt_inst(cli_options.pop("location"), str),
        )


# ########################
# ##### CLICK DECORATORS
# ########################

# These are named as *_options and can be directly applied to click commands/groups as decorators.
# They contain various subsets from the generate_*


def run_config_option(*, name: str, command_name: str) -> Callable[[T_Callable], T_Callable]:
    def wrap(f: T_Callable) -> T_Callable:
        return _apply_click_params(f, _generate_run_config_option(name, command_name))

    return wrap


def job_name_option(f: Optional[T_Callable] = None, *, name: str) -> T_Callable:
    if f is None:
        return lambda f: job_name_option(f, name=name)  # type: ignore
    else:
        return _apply_click_params(f, _generate_job_name_option(name))


def repository_name_option(f: Optional[T_Callable] = None, *, name: str) -> T_Callable:
    if f is None:
        return lambda f: repository_name_option(f, name=name)  # type: ignore
    else:
        return _apply_click_params(f, _generate_repository_name_option(name))


def workspace_options(f: T_Callable) -> T_Callable:
    return _apply_click_params(f, *_generate_workspace_options())


def python_pointer_options(f: T_Callable) -> T_Callable:
    return _apply_click_params(f, *_generate_python_pointer_options(allow_multiple=False))


def repository_options(f: T_Callable) -> T_Callable:
    return _apply_click_params(
        f,
        _generate_repository_name_option("repository"),
        _generate_code_location_name_option("location"),
    )


ClickOption: TypeAlias = Callable[[T_Callable], T_Callable]


def _apply_click_params(command: T_Callable, *click_params: ClickOption) -> T_Callable:
    for click_param in click_params:
        command = click_param(command)
    return command


# ########################
# ##### OPTION GENERATORS
# ########################

# These are named as generate_*_option(s) and return a ClickOption or list of Click Options. These
# cannot be directly applied to click commands/groups as decorators. They are intended to be private
# to this module-- external code should use the below decorators.


def _generate_job_name_option(name: str) -> ClickOption:
    return click.option(
        "--job",
        "-j",
        name,
        help="Job within the repository, necessary if more than one job is present.",
    )


def _generate_repository_name_option(name: str) -> ClickOption:
    return click.option(
        "--repository",
        "-r",
        name,
        help=(
            "Name of the repository, necessary if more than one repository is present in the"
            " code location."
        ),
    )


def _generate_code_location_name_option(name: str) -> ClickOption:
    return click.option(
        "--location",
        "-l",
        name,
        help="Name of the code location, necessary if more than one location is present.",
    )


def _generate_run_config_option(name: str, command_name: str) -> ClickOption:
    return click.option(
        "-c",
        "--config",
        name,
        type=click.Path(exists=True),
        multiple=True,
        help=(
            "Specify one or more run config files. These can also be file patterns. "
            "If more than one run config file is captured then those files are merged. "
            "Files listed first take precedence. They will smash the values of subsequent "
            "files at the key-level granularity. If the file is a pattern then you must "
            "enclose it in double quotes"
            "\n\nExample: "
            f"dagster job {command_name} -f hello_world.py -j pandas_hello_world "
            '-c "pandas_hello_world/*.yaml"'
            "\n\nYou can also specify multiple files:"
            "\n\nExample: "
            f"dagster job {command_name} -f hello_world.py -j pandas_hello_world "
            "-c pandas_hello_world/ops.yaml -c pandas_hello_world/env.yaml"
        ),
    )


def _generate_python_pointer_options(allow_multiple: bool) -> Sequence[ClickOption]:
    return [
        click.option(
            "--working-directory",
            "-d",
            help="Specify working directory to use when loading the repository or job",
            envvar="DAGSTER_WORKING_DIRECTORY",
        ),
        click.option(
            "--python-file",
            "-f",
            # Checks that the path actually exists lower in the stack, where we
            # are better equipped to surface errors
            type=click.Path(exists=False),
            multiple=allow_multiple,
            help=(
                "Specify python file "
                + ("or files (flag can be used multiple times) " if allow_multiple else "")
                + "where dagster definitions reside as top-level symbols/variables and load "
                + ("each" if allow_multiple else "the")
                + " file as a code location in the current python environment."
            ),
            envvar="DAGSTER_PYTHON_FILE",
        ),
        click.option(
            "--module-name",
            "-m",
            multiple=allow_multiple,
            help=(
                "Specify module "
                + ("or modules (flag can be used multiple times) " if allow_multiple else "")
                + "where dagster definitions reside as top-level symbols/variables and load "
                + ("each" if allow_multiple else "the")
                + " module as a code location in the current python environment."
            ),
            envvar="DAGSTER_MODULE_NAME",
        ),
        click.option(
            "--package-name",
            multiple=allow_multiple,
            help="Specify Python package where repository or job function lives",
            envvar="DAGSTER_PACKAGE_NAME",
        ),
        click.option(
            "--attribute",
            "-a",
            help=(
                "Attribute that is either a 1) repository or job or "
                "2) a function that returns a repository or job"
            ),
            envvar="DAGSTER_ATTRIBUTE",
        ),
    ]


def _generate_grpc_server_options(hidden=False) -> Sequence[ClickOption]:
    return [
        click.option(
            "--grpc-port",
            type=click.INT,
            required=False,
            help="Port to use to connect to gRPC server",
            hidden=hidden,
        ),
        click.option(
            "--grpc-socket",
            type=click.Path(),
            required=False,
            help="Named socket to use to connect to gRPC server",
            hidden=hidden,
        ),
        click.option(
            "--grpc-host",
            type=click.STRING,
            required=False,
            help="Host to use to connect to gRPC server, defaults to localhost",
            hidden=hidden,
        ),
        click.option(
            "--use-ssl",
            is_flag=True,
            default=False,
            help="Use a secure channel when connecting to the gRPC server",
            hidden=hidden,
        ),
    ]


def _generate_workspace_options() -> Sequence[ClickOption]:
    return [
        click.option("--empty-workspace", is_flag=True, help="Allow an empty workspace"),
        click.option(
            "--workspace",
            "-w",
            multiple=True,
            type=click.Path(exists=True),
            help="Path to workspace file. Argument can be provided multiple times.",
        ),
        *_generate_python_pointer_options(allow_multiple=True),
        *_generate_grpc_server_options(),
    ]


def _get_code_pointer_dict_from_python_pointer_opts(
    params: PythonPointerOpts,
) -> Mapping[str, CodePointer]:
    working_directory = params.working_directory or os.getcwd()
    loadable_targets = get_loadable_targets(
        params.python_file,
        params.module_name,
        params.package_name,
        working_directory,
        params.attribute,
    )

    # repository_name -> code_pointer
    code_pointer_dict: dict[str, CodePointer] = {}
    for loadable_target in loadable_targets:
        repo_def = check.not_none(repository_def_from_target_def(loadable_target.target_definition))
        if params.python_file:
            code_pointer = CodePointer.from_python_file(
                params.python_file, loadable_target.attribute, working_directory
            )
        elif params.module_name:
            code_pointer = CodePointer.from_module(
                params.module_name, loadable_target.attribute, working_directory
            )
        elif params.package_name:
            code_pointer = CodePointer.from_python_package(
                params.package_name, loadable_target.attribute, working_directory
            )
        else:
            check.failed("Must specify a Python file or module name")

        code_pointer_dict[repo_def.name] = code_pointer

    return code_pointer_dict


def get_repository_python_origin_from_cli_opts(
    params: PythonPointerOpts, repo_name: Optional[str] = None
) -> RepositoryPythonOrigin:
    if sum([bool(x) for x in (params.python_file, params.module_name, params.package_name)]) != 1:
        _raise_cli_usage_error()

    # Short-circuit the case where an attribute and no repository name is passed in,
    # giving us enough information to return an origin without loading any target
    # definitions - we may need to return an origin for a non-existent repository
    # (e.g. to log an origin ID for an error message)
    if params.attribute and not repo_name:
        working_directory = params.working_directory or os.getcwd()
        if params.python_file:
            code_pointer: CodePointer = CodePointer.from_python_file(
                params.python_file,
                params.attribute,
                working_directory,
            )
        elif params.module_name:
            code_pointer = CodePointer.from_module(
                params.module_name,
                params.attribute,
                working_directory,
            )
        elif params.package_name:
            code_pointer = CodePointer.from_python_package(
                params.package_name,
                params.attribute,
                working_directory,
            )
        else:
            check.failed("Must specify a Python file or module name")
        return RepositoryPythonOrigin(
            executable_path=sys.executable,
            code_pointer=code_pointer,
            entry_point=DEFAULT_DAGSTER_ENTRY_POINT,
        )

    code_pointer_dict = _get_code_pointer_dict_from_python_pointer_opts(params)
    found_repo_names = serialize_sorted_quoted(code_pointer_dict.keys())
    if repo_name is None and len(code_pointer_dict) == 1:
        code_pointer = next(iter(code_pointer_dict.values()))
    elif repo_name is None:
        raise click.UsageError(
            "Must provide --repository as there is more than one repository. "
            f"Options are: {found_repo_names}."
        )
    elif repo_name not in code_pointer_dict:
        raise click.UsageError(
            f'Repository "{repo_name}" not found. Found {found_repo_names} instead.'
        )
    else:
        code_pointer = code_pointer_dict[repo_name]

    return RepositoryPythonOrigin(
        executable_path=sys.executable,
        code_pointer=code_pointer,
        entry_point=DEFAULT_DAGSTER_ENTRY_POINT,
    )


def get_code_location_from_workspace(
    workspace: WorkspaceRequestContext, code_location_name: Optional[str]
) -> CodeLocation:
    if code_location_name is None:
        if len(workspace.code_location_names) == 1:
            code_location_name = workspace.code_location_names[0]
        elif len(workspace.code_location_names) == 0:
            raise click.UsageError("No locations found in workspace")
        elif code_location_name is None:
            raise click.UsageError(
                "Must provide --location as there are multiple locations "
                f"available. Options are: {serialize_sorted_quoted(workspace.code_location_names)}"
            )

    if code_location_name not in workspace.code_location_names:
        raise click.UsageError(
            f'Location "{code_location_name}" not found in workspace. '
            f"Found {serialize_sorted_quoted(workspace.code_location_names)} instead."
        )

    if workspace.has_code_location_error(code_location_name):
        raise click.UsageError(
            f'Error loading location "{code_location_name}": {workspace.get_code_location_error(code_location_name)!s}'
        )

    return workspace.get_code_location(code_location_name)


def get_remote_repository_from_code_location(
    code_location: CodeLocation, provided_repo_name: Optional[str]
) -> RemoteRepository:
    check.inst_param(code_location, "code_location", CodeLocation)
    check.opt_str_param(provided_repo_name, "provided_repo_name")

    repo_dict = code_location.get_repositories()
    check.invariant(repo_dict, "There should be at least one repo.")

    # no name provided and there is only one repo. Automatically return
    if provided_repo_name is None and len(repo_dict) == 1:
        return next(iter(repo_dict.values()))

    if provided_repo_name is None:
        raise click.UsageError(
            "Must provide --repository as there is more than one repository "
            f"in {code_location.name}. Options are: {serialize_sorted_quoted(repo_dict.keys())}."
        )

    if not code_location.has_repository(provided_repo_name):
        raise click.UsageError(
            f'Repository "{provided_repo_name}" not found in location "{code_location.name}". '
            f"Found {serialize_sorted_quoted(repo_dict.keys())} instead."
        )

    return code_location.get_repository(provided_repo_name)


def get_remote_job_from_remote_repo(
    remote_repo: RemoteRepository,
    provided_name: Optional[str],
) -> RemoteJob:
    check.inst_param(remote_repo, "remote_repo", RemoteRepository)
    check.opt_str_param(provided_name, "provided_name")

    remote_jobs = {ep.name: ep for ep in (remote_repo.get_all_jobs())}

    check.invariant(remote_jobs)

    if provided_name is None and len(remote_jobs) == 1:
        return next(iter(remote_jobs.values()))

    if provided_name is None:
        raise click.UsageError(
            "Must provide --job as there is more than one job "
            f"in {remote_repo.name}. Options are: {serialize_sorted_quoted(remote_jobs.keys())}."
        )

    if provided_name not in remote_jobs:
        raise click.UsageError(
            f'Job "{provided_name}" not found in repository "{remote_repo.name}". '
            f"Found {serialize_sorted_quoted(remote_jobs.keys())} instead."
        )

    return remote_jobs[provided_name]


def get_run_config_from_file_list(file_list: list[str]) -> Mapping[str, object]:
    check.opt_sequence_param(file_list, "file_list", of_type=str)
    return cast(Mapping[str, object], load_yaml_from_glob_list(file_list) if file_list else {})


def get_run_config_from_cli_opts(
    config_files: tuple[str, ...], config_json: Optional[str]
) -> Mapping[str, object]:
    if not (config_files or config_json):
        return {}
    elif config_files and config_json:
        raise click.UsageError("Cannot specify both -c / --config and --config-json")
    elif config_files:
        return get_run_config_from_file_list(list(config_files))
    elif config_json:
        try:
            return json.loads(config_json)
        except JSONDecodeError:
            raise click.UsageError(
                f"Invalid JSON-string given for `--config-json`: {config_json}\n\n{serializable_error_info_from_exc_info(sys.exc_info()).to_string()}"
            )
    else:
        check.failed("Unhandled case getting config from kwargs")


# ########################
# ##### HELPERS
# ########################


def _raise_cli_usage_error(msg: Optional[str] = None) -> Never:
    raise UsageError(
        msg or "Invalid set of CLI arguments for loading repository/job. See --help for details."
    )


def _check_attrs_falsey(obj: object, *attrs: str) -> None:
    if not _are_attrs_falsey(obj, *attrs):
        _raise_cli_usage_error()


def _are_attrs_falsey(obj: object, *attrs: str) -> bool:
    for attr in attrs:
        if getattr(obj, attr):
            return False
    return True



import os
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import NamedTuple, Optional, Union, cast

import tomli

from dagster._core.remote_representation.origin import (
    CodeLocationOrigin,
    GrpcServerCodeLocationOrigin,
    InProcessCodeLocationOrigin,
    ManagedGrpcPythonEnvCodeLocationOrigin,
)
from dagster._core.workspace.load import (
    location_origin_from_module_name,
    location_origin_from_package_name,
    location_origin_from_python_file,
    location_origins_from_yaml_paths,
)


class WorkspaceLoadTarget(ABC):
    @abstractmethod
    def create_origins(self) -> Sequence[CodeLocationOrigin]:
        """Reloads the CodeLocationOrigins for this workspace."""


class CompositeTarget(
    NamedTuple("CompositeTarget", [("targets", Sequence[WorkspaceLoadTarget])]),
    WorkspaceLoadTarget,
):
    def create_origins(self):
        origins = []
        for target in self.targets:
            origins += target.create_origins()
        return origins


class WorkspaceFileTarget(
    NamedTuple("WorkspaceFileTarget", [("paths", Sequence[str])]), WorkspaceLoadTarget
):
    def create_origins(self) -> Sequence[CodeLocationOrigin]:
        return location_origins_from_yaml_paths(self.paths)


class InProcessWorkspaceLoadTarget(WorkspaceLoadTarget):
    """A workspace load target that is in-process and does not spin up a gRPC server."""

    def __init__(
        self, origin: Union[InProcessCodeLocationOrigin, Sequence[InProcessCodeLocationOrigin]]
    ):
        self._origins = cast(
            Sequence[InProcessCodeLocationOrigin],
            origin if isinstance(origin, list) else [origin],
        )

    def create_origins(self) -> Sequence[InProcessCodeLocationOrigin]:
        return self._origins


def validate_dagster_block_for_module_name_or_modules(dagster_block):
    module_name_present = "module_name" in dagster_block and isinstance(
        dagster_block.get("module_name"), str
    )
    modules_present = "modules" in dagster_block and isinstance(dagster_block.get("modules"), list)

    if module_name_present and modules_present:
        # Here we have the check only for list; to be a bit more forgiving in comparison to 'is_valid_modules_list' in case it's an empty list next to 'module_name' existance
        if len(dagster_block["modules"]) > 0:
            raise ValueError(
                "Only one of 'module_name' or 'modules' should be specified, not both."
            )

    if modules_present and len(dagster_block.get("modules")) == 0:
        raise ValueError("'modules' list should not be empty if specified.")

    return True


def is_valid_modules_list(modules: list[dict[str, str]]) -> bool:
    # Could be skipped theorectically, but double check maybe useful, if this functions finds it's way elsewhere
    if not isinstance(modules, list):
        raise ValueError("Modules should be a list.")

    for index, item in enumerate(modules):
        if not isinstance(item, dict):
            raise ValueError(f"Item at index {index} is not a dictionary.")
        if "type" not in item:
            raise ValueError(f"Dictionary at index {index} does not contain the key 'type'.")
        if not isinstance(item["type"], str):
            raise ValueError(f"The 'type' value in dictionary at index {index} is not a string.")
        if "name" not in item:
            raise ValueError(f"Dictionary at index {index} does not contain the key 'name'.")
        if not isinstance(item["name"], str):
            raise ValueError(f"The 'name' value in dictionary at index {index} is not a string.")

    return True


def get_origins_from_toml(
    path: str,
) -> Sequence[ManagedGrpcPythonEnvCodeLocationOrigin]:
    with open(path, "rb") as f:
        data = tomli.load(f)
        if not isinstance(data, dict):
            return []

        dagster_block = data.get("tool", {}).get("dagster", {})

        if "module_name" in dagster_block or "modules" in dagster_block:
            assert validate_dagster_block_for_module_name_or_modules(dagster_block) is True

        if "module_name" in dagster_block:
            return ModuleTarget(
                module_name=dagster_block.get("module_name"),
                attribute=None,
                working_directory=os.getcwd(),
                location_name=dagster_block.get("code_location_name"),
            ).create_origins()
        elif "modules" in dagster_block and is_valid_modules_list(dagster_block.get("modules")):
            origins = []
            for module in dagster_block.get("modules"):
                if module.get("type") == "module":
                    origins.extend(
                        ModuleTarget(
                            module_name=module.get("name"),
                            attribute=None,
                            working_directory=os.getcwd(),
                            location_name=dagster_block.get("code_location_name"),
                        ).create_origins()
                    )
            return origins
        else:
            return []


class PyProjectFileTarget(NamedTuple("PyProjectFileTarget", [("path", str)]), WorkspaceLoadTarget):
    def create_origins(self) -> Sequence[CodeLocationOrigin]:
        return get_origins_from_toml(self.path)


class PythonFileTarget(
    NamedTuple(
        "PythonFileTarget",
        [
            ("python_file", str),
            ("attribute", Optional[str]),
            ("working_directory", Optional[str]),
            ("location_name", Optional[str]),
        ],
    ),
    WorkspaceLoadTarget,
):
    def create_origins(self) -> Sequence[ManagedGrpcPythonEnvCodeLocationOrigin]:
        return [
            location_origin_from_python_file(
                python_file=self.python_file,
                attribute=self.attribute,
                working_directory=self.working_directory,
                location_name=self.location_name,
            )
        ]


class ModuleTarget(
    NamedTuple(
        "ModuleTarget",
        [
            ("module_name", str),
            ("attribute", Optional[str]),
            ("working_directory", Optional[str]),
            ("location_name", Optional[str]),
        ],
    ),
    WorkspaceLoadTarget,
):
    def create_origins(self) -> Sequence[ManagedGrpcPythonEnvCodeLocationOrigin]:
        return [
            location_origin_from_module_name(
                self.module_name,
                self.attribute,
                self.working_directory,
                location_name=self.location_name,
            )
        ]


class PackageTarget(
    NamedTuple(
        "ModuleTarget",
        [
            ("package_name", str),
            ("attribute", Optional[str]),
            ("working_directory", Optional[str]),
            ("location_name", Optional[str]),
        ],
    ),
    WorkspaceLoadTarget,
):
    def create_origins(self) -> Sequence[ManagedGrpcPythonEnvCodeLocationOrigin]:
        return [
            location_origin_from_package_name(
                self.package_name,
                self.attribute,
                self.working_directory,
                location_name=self.location_name,
            )
        ]


class GrpcServerTarget(
    NamedTuple(
        "ModuleTarget",
        [
            ("host", str),
            ("port", Optional[int]),
            ("socket", Optional[str]),
            ("location_name", Optional[str]),
        ],
    ),
    WorkspaceLoadTarget,
):
    def create_origins(self) -> Sequence[GrpcServerCodeLocationOrigin]:
        return [
            GrpcServerCodeLocationOrigin(
                port=self.port,
                socket=self.socket,
                host=self.host,
                location_name=self.location_name,
            )
        ]


#  Utility target for graphql commands that do not require a workspace, e.g. downloading schema
class EmptyWorkspaceTarget(NamedTuple("EmptyWorkspaceTarget", []), WorkspaceLoadTarget):
    def create_origins(self) -> Sequence[CodeLocationOrigin]:
        return []



import logging
import sys
from collections.abc import Iterator, Sequence
from typing import Optional, cast

import dagster._check as check
from dagster._core.definitions.metadata import MetadataValue
from dagster._core.definitions.selector import JobSubsetSelector
from dagster._core.events import EngineEventData, RunFailureReason
from dagster._core.execution.plan.resume_retry import ReexecutionStrategy
from dagster._core.execution.retries import auto_reexecution_should_retry_run
from dagster._core.instance import DagsterInstance
from dagster._core.storage.dagster_run import DagsterRun, DagsterRunStatus, RunRecord
from dagster._core.storage.tags import (
    AUTO_RETRY_RUN_ID_TAG,
    RETRY_NUMBER_TAG,
    RETRY_ON_ASSET_OR_OP_FAILURE_TAG,
    RETRY_STRATEGY_TAG,
    RUN_FAILURE_REASON_TAG,
    WILL_RETRY_TAG,
)
from dagster._core.workspace.context import IWorkspaceProcessContext
from dagster._daemon.utils import DaemonErrorCapture
from dagster._utils.tags import get_boolean_tag_value

DEFAULT_REEXECUTION_POLICY = ReexecutionStrategy.FROM_FAILURE


def should_retry(run: DagsterRun, instance: DagsterInstance) -> bool:
    """A more robust method of determining is a run should be retried by the daemon than just looking
    at the WILL_RETRY_TAG. We account for the case where the code version is old and doesn't set the
    WILL_RETRY_TAG. If the tag wasn't set for a run failure, we set it so that other daemons can use the
    WILL_RETRY_TAG to determine if the run should be retried.
    """
    will_retry_tag_value = run.tags.get(WILL_RETRY_TAG)
    run_failure_reason = (
        RunFailureReason(run.tags.get(RUN_FAILURE_REASON_TAG))
        if run.tags.get(RUN_FAILURE_REASON_TAG)
        else None
    )
    if will_retry_tag_value is None:
        # If the run doesn't have the WILL_RETRY_TAG, and the run is failed, we
        # recalculate if the run should be retried to ensure backward compatibilty
        if run.status == DagsterRunStatus.FAILURE:
            should_retry_run = auto_reexecution_should_retry_run(instance, run, run_failure_reason)
            # add the tag to the run so that it can be used in other parts of the system
            instance.add_run_tags(run.run_id, {WILL_RETRY_TAG: str(should_retry_run).lower()})
        else:
            # run is not failed, and shouldn't be retried
            return False
    else:
        should_retry_run = get_boolean_tag_value(will_retry_tag_value, default_value=False)

    if should_retry_run:
        return should_retry_run
    else:
        # one of the reasons we may not retry a run is if it is a step failure and system is
        # set to not retry on op/asset failures. In this case, we log
        # an engine event
        retry_on_asset_or_op_failure = get_boolean_tag_value(
            run.tags.get(RETRY_ON_ASSET_OR_OP_FAILURE_TAG),
            default_value=instance.run_retries_retry_on_asset_or_op_failure,
        )
        if run_failure_reason == RunFailureReason.STEP_FAILURE and not retry_on_asset_or_op_failure:
            instance.report_engine_event(
                "Not retrying run since it failed due to an asset or op failure and run retries "
                "are configured with retry_on_asset_or_op_failure set to false.",
                run,
            )
        return False


def filter_runs_to_should_retry(
    runs: Sequence[DagsterRun], instance: DagsterInstance
) -> Iterator[DagsterRun]:
    """Return only runs that should retry along with their retry number (1st retry, 2nd, etc.)."""
    for run in runs:
        if should_retry(run, instance):
            yield run


def get_automatically_retried_run_if_exists(
    instance: DagsterInstance, run: DagsterRun, run_group: Sequence[DagsterRun]
) -> Optional[DagsterRun]:
    if run.tags.get(AUTO_RETRY_RUN_ID_TAG) is not None:
        return instance.get_run_by_id(run.tags[AUTO_RETRY_RUN_ID_TAG])
    child_run = next(
        (retried_run for retried_run in run_group if run.run_id == retried_run.parent_run_id), None
    )
    if child_run is not None and child_run.tags.get(RETRY_NUMBER_TAG) is not None:
        # We use the presense of RETRY_NUMBER_TAG to confirm that the child run was launched
        # by the automatic retry daemon. If the child run was launched by the user, the tag
        # should not be present.
        return child_run


def get_reexecution_strategy(
    run: DagsterRun, instance: DagsterInstance
) -> Optional[ReexecutionStrategy]:
    raw_strategy_tag = run.tags.get(RETRY_STRATEGY_TAG)
    if raw_strategy_tag is None:
        return None

    if raw_strategy_tag not in ReexecutionStrategy.__members__:
        instance.report_engine_event(
            f"Error parsing retry strategy from tag '{RETRY_STRATEGY_TAG}: {raw_strategy_tag}'", run
        )
        return None
    else:
        return ReexecutionStrategy[raw_strategy_tag]


def retry_run(
    failed_run: DagsterRun,
    workspace_context: IWorkspaceProcessContext,
) -> None:
    """Submit a retry as a re-execute from failure."""
    instance = workspace_context.instance
    workspace = workspace_context.create_request_context()
    if not failed_run.remote_job_origin:
        instance.report_engine_event(
            "Run does not have an external job origin, unable to retry the run.",
            failed_run,
        )
        return

    origin = failed_run.remote_job_origin.repository_origin
    code_location = workspace.get_code_location(origin.code_location_origin.location_name)
    repo_name = origin.repository_name

    if not code_location.has_repository(repo_name):
        instance.report_engine_event(
            f"Could not find repository {repo_name} in location {code_location.name}, unable to"
            " retry the run. It was likely renamed or deleted.",
            failed_run,
        )
        return

    repo = code_location.get_repository(repo_name)

    if not repo.has_job(failed_run.job_name):
        instance.report_engine_event(
            f"Could not find job {failed_run.job_name} in repository {repo_name}, unable"
            " to retry the run. It was likely renamed or deleted.",
            failed_run,
        )
        return

    remote_job = code_location.get_job(
        JobSubsetSelector(
            location_name=origin.code_location_origin.location_name,
            repository_name=repo_name,
            job_name=failed_run.job_name,
            op_selection=failed_run.op_selection,
            asset_selection=(
                None if failed_run.asset_selection is None else list(failed_run.asset_selection)
            ),
        )
    )

    _, run_group = check.not_none(instance.get_run_group(failed_run.run_id))
    run_group_list = list(run_group)

    # it is possible for the daemon to die between creating the run and submitting it. We account for this
    # possibility by checking if the a run already exists in the run group with the parent run id of the
    # failed run and resubmit it if necessary.
    existing_retried_run = get_automatically_retried_run_if_exists(
        instance=instance, run=failed_run, run_group=run_group_list
    )
    if existing_retried_run is not None:
        # ensure the failed_run has the AUTO_RETRY_RUN_ID_TAG set
        if failed_run.tags.get(AUTO_RETRY_RUN_ID_TAG) is None:
            instance.add_run_tags(
                failed_run.run_id, {AUTO_RETRY_RUN_ID_TAG: existing_retried_run.run_id}
            )
        if existing_retried_run.status == DagsterRunStatus.NOT_STARTED:
            # A run already exists but was not submitted.
            instance.submit_run(existing_retried_run.run_id, workspace)
        return

    # At this point we know we need to launch a new run for the retry
    strategy = get_reexecution_strategy(failed_run, instance) or DEFAULT_REEXECUTION_POLICY
    tags = {RETRY_NUMBER_TAG: str(len(run_group_list))}
    new_run = instance.create_reexecuted_run(
        parent_run=failed_run,
        code_location=code_location,
        remote_job=remote_job,
        strategy=strategy,
        extra_tags=tags,
        use_parent_run_tags=True,
    )
    instance.add_run_tags(failed_run.run_id, {AUTO_RETRY_RUN_ID_TAG: new_run.run_id})

    instance.report_engine_event(
        "Retrying the run",
        failed_run,
        engine_event_data=EngineEventData({"new run": MetadataValue.dagster_run(new_run.run_id)}),
    )
    instance.report_engine_event(
        "Launched as an automatic retry",
        new_run,
        engine_event_data=EngineEventData(
            {"failed run": MetadataValue.dagster_run(failed_run.run_id)}
        ),
    )

    instance.submit_run(new_run.run_id, workspace)


def consume_new_runs_for_automatic_reexecution(
    workspace_process_context: IWorkspaceProcessContext,
    run_records: Sequence[RunRecord],
    logger: logging.Logger,
) -> Iterator[None]:
    """Check which runs should be retried, and retry them.

    It's safe to call this method on the same run multiple times because once a retry run is created,
    it won't create another. The only exception is if the new run gets deleted, in which case we'd
    retry the run again.
    """
    for run in filter_runs_to_should_retry(
        [cast(DagsterRun, run_record.dagster_run) for run_record in run_records],
        workspace_process_context.instance,
    ):
        yield
        try:
            retry_run(run, workspace_process_context)
        except Exception:
            error_info = DaemonErrorCapture.process_exception(
                exc_info=sys.exc_info(),
                logger=logger,
                log_message=f"Failed to retry run {run.run_id}",
            )
            workspace_process_context.instance.report_engine_event(
                "Failed to retry run",
                run,
                engine_event_data=EngineEventData(error=error_info),
            )
            # Since something failed when retrying this run, mark that we will not retry it so that we
            # don't retry it again in the future, and so that the tags reflect the state of the system.
            # We may want to split out the kinds of exceptions and handle them differently in the future so
            # that this can be more resiliant to transient errors. However, this would also require some changes
            # to the EventLogConsumerDaemon so that the cursors are not updated in a way that prevents this run
            # from being processed in the next tick.
            workspace_process_context.instance.add_run_tags(
                run_id=run.run_id, new_tags={WILL_RETRY_TAG: "false"}
            )



from collections.abc import Sequence
from typing import TYPE_CHECKING, TypeVar, Union

from typing_extensions import TypeAlias

from dagster._core.definitions.graph_definition import GraphDefinition
from dagster._core.definitions.job_definition import JobDefinition
from dagster._core.definitions.schedule_definition import ScheduleDefinition
from dagster._core.definitions.sensor_definition import SensorDefinition
from dagster._core.definitions.source_asset import SourceAsset
from dagster._core.definitions.unresolved_asset_job_definition import UnresolvedAssetJobDefinition

if TYPE_CHECKING:
    from dagster._core.definitions import AssetsDefinition
    from dagster._core.definitions.cacheable_assets import CacheableAssetsDefinition
    from dagster._core.definitions.partitioned_schedule import (
        UnresolvedPartitionedAssetScheduleDefinition,
    )

SINGLETON_REPOSITORY_NAME = "__repository__"

VALID_REPOSITORY_DATA_DICT_KEYS = {
    "schedules",
    "sensors",
    "jobs",
}

T_RepositoryLevelDefinition = TypeVar(
    "T_RepositoryLevelDefinition",
    JobDefinition,
    ScheduleDefinition,
    SensorDefinition,
)

RepositoryElementDefinition: TypeAlias = Union[
    "AssetsDefinition",
    GraphDefinition,
    JobDefinition,
    ScheduleDefinition,
    SensorDefinition,
    SourceAsset,
    UnresolvedAssetJobDefinition,
    "UnresolvedPartitionedAssetScheduleDefinition",
]

RepositoryDictSpec: TypeAlias = dict[str, dict[str, RepositoryElementDefinition]]

RepositoryListSpec: TypeAlias = Sequence[
    Union[RepositoryElementDefinition, "CacheableAssetsDefinition"]
]



from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping, Sequence
from typing import TYPE_CHECKING, AbstractSet, NamedTuple, Optional, Union  # noqa: UP035

import dagster._check as check
from dagster._core.assets import AssetDetails
from dagster._core.definitions.asset_check_spec import AssetCheckKey
from dagster._core.definitions.data_version import DATA_VERSION_TAG
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.partition import PartitionsDefinition
from dagster._core.event_api import (
    AssetRecordsFilter,
    EventHandlerFn,
    EventLogCursor,
    EventLogRecord,
    EventRecordsFilter,
    EventRecordsResult,
    RunStatusChangeRecordsFilter,
)
from dagster._core.events import DagsterEventType
from dagster._core.execution.stats import (
    RUN_STATS_EVENT_TYPES,
    STEP_STATS_EVENT_TYPES,
    RunStepKeyStatsSnapshot,
    build_run_stats_from_events,
    build_run_step_stats_from_events,
)
from dagster._core.instance import MayHaveInstanceWeakref, T_DagsterInstance
from dagster._core.instance.config import PoolConfig
from dagster._core.loader import LoadableBy, LoadingContext
from dagster._core.storage.asset_check_execution_record import (
    AssetCheckExecutionRecord,
    AssetCheckExecutionRecordStatus,
)
from dagster._core.storage.dagster_run import DagsterRunStatsSnapshot
from dagster._core.storage.partition_status_cache import get_and_update_asset_status_cache_value
from dagster._core.storage.sql import AlembicVersion
from dagster._core.storage.tags import MULTIDIMENSIONAL_PARTITION_PREFIX
from dagster._record import record
from dagster._utils import PrintFn
from dagster._utils.concurrency import ConcurrencyClaimStatus, ConcurrencyKeyInfo
from dagster._utils.warnings import deprecation_warning

if TYPE_CHECKING:
    from dagster._core.events.log import EventLogEntry
    from dagster._core.storage.partition_status_cache import AssetStatusCacheValue


class EventLogConnection(NamedTuple):
    records: Sequence[EventLogRecord]
    cursor: str
    has_more: bool


class AssetEntry(
    NamedTuple(
        "_AssetEntry",
        [
            ("asset_key", AssetKey),
            ("last_materialization_record", Optional[EventLogRecord]),
            ("last_run_id", Optional[str]),
            ("asset_details", Optional[AssetDetails]),
            ("cached_status", Optional["AssetStatusCacheValue"]),
            # This is an optional field which can be used for more performant last observation
            # queries if the underlying storage supports it
            ("last_observation_record", Optional[EventLogRecord]),
            ("last_planned_materialization_storage_id", Optional[int]),
            ("last_planned_materialization_run_id", Optional[str]),
        ],
    )
):
    def __new__(
        cls,
        asset_key: AssetKey,
        last_materialization_record: Optional[EventLogRecord] = None,
        last_run_id: Optional[str] = None,
        asset_details: Optional[AssetDetails] = None,
        cached_status: Optional["AssetStatusCacheValue"] = None,
        last_observation_record: Optional[EventLogRecord] = None,
        last_planned_materialization_storage_id: Optional[int] = None,
        last_planned_materialization_run_id: Optional[str] = None,
    ):
        from dagster._core.storage.partition_status_cache import AssetStatusCacheValue

        return super().__new__(
            cls,
            asset_key=check.inst_param(asset_key, "asset_key", AssetKey),
            last_materialization_record=check.opt_inst_param(
                last_materialization_record,
                "last_materialization_record",
                EventLogRecord,
            ),
            last_run_id=check.opt_str_param(last_run_id, "last_run_id"),
            asset_details=check.opt_inst_param(asset_details, "asset_details", AssetDetails),
            cached_status=check.opt_inst_param(
                cached_status, "cached_status", AssetStatusCacheValue
            ),
            last_observation_record=check.opt_inst_param(
                last_observation_record, "last_observation_record", EventLogRecord
            ),
            last_planned_materialization_storage_id=check.opt_int_param(
                last_planned_materialization_storage_id,
                "last_planned_materialization_storage_id",
            ),
            last_planned_materialization_run_id=check.opt_str_param(
                last_planned_materialization_run_id,
                "last_planned_materialization_run_id",
            ),
        )

    @property
    def last_materialization(self) -> Optional["EventLogEntry"]:
        if self.last_materialization_record is None:
            return None
        return self.last_materialization_record.event_log_entry

    @property
    def last_observation(self) -> Optional["EventLogEntry"]:
        if self.last_observation_record is None:
            return None
        return self.last_observation_record.event_log_entry

    @property
    def last_materialization_storage_id(self) -> Optional[int]:
        if self.last_materialization_record is None:
            return None
        return self.last_materialization_record.storage_id


class AssetRecord(
    NamedTuple("_NamedTuple", [("storage_id", int), ("asset_entry", AssetEntry)]),
    LoadableBy[AssetKey],
):
    """Internal representation of an asset record, as stored in a :py:class:`~dagster._core.storage.event_log.EventLogStorage`.

    Users should not invoke this class directly.
    """

    @classmethod
    def _blocking_batch_load(
        cls, keys: Iterable[AssetKey], context: LoadingContext
    ) -> Iterable[Optional["AssetRecord"]]:
        records_by_key = {
            record.asset_entry.asset_key: record
            for record in context.instance.get_asset_records(list(keys))
        }
        return [records_by_key.get(key) for key in keys]


class AssetCheckSummaryRecord(
    NamedTuple(
        "_AssetCheckSummaryRecord",
        [
            ("asset_check_key", AssetCheckKey),
            ("last_check_execution_record", Optional[AssetCheckExecutionRecord]),
            ("last_run_id", Optional[str]),
            ("last_completed_check_execution_record", Optional[AssetCheckExecutionRecord]),
        ],
    ),
    LoadableBy[AssetCheckKey],
):
    @classmethod
    def _blocking_batch_load(
        cls, keys: Iterable[AssetCheckKey], context: LoadingContext
    ) -> Iterable[Optional["AssetCheckSummaryRecord"]]:
        records_by_key = context.instance.event_log_storage.get_asset_check_summary_records(
            list(keys)
        )
        return [records_by_key[key] for key in keys]

    @property
    def last_completed_run_id(self) -> Optional[str]:
        return (
            self.last_completed_check_execution_record.run_id
            if self.last_completed_check_execution_record
            else None
        )


class PlannedMaterializationInfo(NamedTuple):
    """Internal representation of an planned materialization event, containing storage_id / run_id.

    Users should not invoke this class directly.
    """

    storage_id: int
    run_id: str


@record
class PoolLimit:
    name: str
    limit: int
    from_default: bool


class EventLogStorage(ABC, MayHaveInstanceWeakref[T_DagsterInstance]):
    """Abstract base class for storing structured event logs from pipeline runs.

    Note that event log storages using SQL databases as backing stores should implement
    :py:class:`~dagster._core.storage.event_log.SqlEventLogStorage`.

    Users should not directly instantiate concrete subclasses of this class; they are instantiated
    by internal machinery when ``dagster-webserver`` and ``dagster-graphql`` load, based on the values in the
    ``dagster.yaml`` file in ``$DAGSTER_HOME``. Configuration of concrete subclasses of this class
    should be done by setting values in that file.
    """

    def get_logs_for_run(
        self,
        run_id: str,
        cursor: Optional[Union[str, int]] = None,
        of_type: Optional[Union[DagsterEventType, set[DagsterEventType]]] = None,
        limit: Optional[int] = None,
        ascending: bool = True,
    ) -> Sequence["EventLogEntry"]:
        """Get all of the logs corresponding to a run.

        Args:
            run_id (str): The id of the run for which to fetch logs.
            cursor (Optional[Union[str, int]]): Cursor value to track paginated queries.  Legacy
                support for integer offset cursors.
            of_type (Optional[DagsterEventType]): the dagster event type to filter the logs.
            limit (Optional[int]): Max number of records to return.
        """
        if isinstance(cursor, int):
            deprecation_warning(
                "Integer cursor values in `get_logs_for_run`",
                "1.8.0",
                "You can instead construct a storage_id-based string cursor e.g. `cursor = EventLogCursor.from_storage_id(storage_id).to_string()`",
            )
            cursor = EventLogCursor.from_offset(cursor + 1).to_string()
        records = self.get_records_for_run(
            run_id, cursor, of_type, limit, ascending=ascending
        ).records
        return [record.event_log_entry for record in records]

    @abstractmethod
    def get_records_for_run(
        self,
        run_id: str,
        cursor: Optional[str] = None,
        of_type: Optional[Union[DagsterEventType, set[DagsterEventType]]] = None,
        limit: Optional[int] = None,
        ascending: bool = True,
    ) -> EventLogConnection:
        """Get all of the event log records corresponding to a run.

        Args:
            run_id (str): The id of the run for which to fetch logs.
            cursor (Optional[str]): Cursor value to track paginated queries.
            of_type (Optional[DagsterEventType]): the dagster event type to filter the logs.
            limit (Optional[int]): Max number of records to return.
        """

    def get_stats_for_run(self, run_id: str) -> DagsterRunStatsSnapshot:
        """Get a summary of events that have ocurred in a run."""
        return build_run_stats_from_events(
            run_id, self.get_logs_for_run(run_id, of_type=RUN_STATS_EVENT_TYPES)
        )

    def get_step_stats_for_run(
        self, run_id: str, step_keys: Optional[Sequence[str]] = None
    ) -> Sequence[RunStepKeyStatsSnapshot]:
        """Get per-step stats for a pipeline run."""
        logs = self.get_logs_for_run(run_id, of_type=STEP_STATS_EVENT_TYPES)
        if step_keys:
            logs = [
                event
                for event in logs
                if event.is_dagster_event and event.get_dagster_event().step_key in step_keys
            ]

        return build_run_step_stats_from_events(run_id, logs)

    @abstractmethod
    def store_event(self, event: "EventLogEntry") -> None:
        """Store an event corresponding to a pipeline run.

        Args:
            event (EventLogEntry): The event to store.
        """

    def store_event_batch(self, events: Sequence["EventLogEntry"]) -> None:
        for event in events:
            self.store_event(event)

    @abstractmethod
    def delete_events(self, run_id: str) -> None:
        """Remove events for a given run id."""

    @abstractmethod
    def upgrade(self) -> None:
        """This method should perform any schema migrations necessary to bring an
        out-of-date instance of the storage up to date.
        """

    @abstractmethod
    def reindex_events(self, print_fn: Optional[PrintFn] = None, force: bool = False) -> None:
        """Call this method to run any data migrations across the event_log tables."""

    @abstractmethod
    def reindex_assets(self, print_fn: Optional[PrintFn] = None, force: bool = False) -> None:
        """Call this method to run any data migrations across the asset tables."""

    @abstractmethod
    def wipe(self) -> None:
        """Clear the log storage."""

    @abstractmethod
    def watch(self, run_id: str, cursor: Optional[str], callback: EventHandlerFn) -> None:
        """Call this method to start watching."""

    @abstractmethod
    def end_watch(self, run_id: str, handler: EventHandlerFn) -> None:
        """Call this method to stop watching."""

    @property
    @abstractmethod
    def is_persistent(self) -> bool:
        """bool: Whether the storage is persistent."""

    def dispose(self) -> None:
        """Explicit lifecycle management."""

    def optimize_for_webserver(
        self, statement_timeout: int, pool_recycle: int, max_overflow: int
    ) -> None:
        """Allows for optimizing database connection / use in the context of a long lived webserver process."""

    @abstractmethod
    def get_event_records(
        self,
        event_records_filter: EventRecordsFilter,
        limit: Optional[int] = None,
        ascending: bool = False,
    ) -> Sequence[EventLogRecord]:
        pass

    def get_logs_for_all_runs_by_log_id(
        self,
        after_cursor: int = -1,
        dagster_event_type: Optional[Union[DagsterEventType, set[DagsterEventType]]] = None,
        limit: Optional[int] = None,
    ) -> Mapping[int, "EventLogEntry"]:
        """Get event records across all runs. Only supported for non sharded sql storage."""
        raise NotImplementedError()

    def get_maximum_record_id(self) -> Optional[int]:
        """Get the current greatest record id in the event log. Only supported for non sharded sql storage."""
        raise NotImplementedError()

    @abstractmethod
    def can_read_asset_status_cache(self) -> bool:
        """Whether the storage can access cached status information for each asset."""
        pass

    @abstractmethod
    def can_write_asset_status_cache(self) -> bool:
        """Whether the storage is able to write to that cache."""

    @abstractmethod
    def wipe_asset_cached_status(self, asset_key: AssetKey) -> None:
        pass

    @abstractmethod
    def get_asset_records(
        self, asset_keys: Optional[Sequence[AssetKey]] = None
    ) -> Sequence[AssetRecord]:
        pass

    @abstractmethod
    def get_asset_check_summary_records(
        self, asset_check_keys: Sequence[AssetCheckKey]
    ) -> Mapping[AssetCheckKey, AssetCheckSummaryRecord]:
        pass

    @property
    def asset_records_have_last_planned_materialization_storage_id(self) -> bool:
        return False

    @abstractmethod
    def has_asset_key(self, asset_key: AssetKey) -> bool:
        pass

    @abstractmethod
    def all_asset_keys(self) -> Sequence[AssetKey]:
        pass

    @abstractmethod
    def update_asset_cached_status_data(
        self, asset_key: AssetKey, cache_values: "AssetStatusCacheValue"
    ) -> None:
        pass

    def get_asset_keys(
        self,
        prefix: Optional[Sequence[str]] = None,
        limit: Optional[int] = None,
        cursor: Optional[str] = None,
    ) -> Sequence[AssetKey]:
        # base implementation of get_asset_keys, using the existing `all_asset_keys` and doing the
        # filtering in-memory
        asset_keys = sorted(self.all_asset_keys(), key=lambda k: k.to_string())
        if prefix:
            asset_keys = [
                asset_key for asset_key in asset_keys if asset_key.path[: len(prefix)] == prefix
            ]
        if cursor:
            cursor_asset = AssetKey.from_db_string(cursor)

            if not cursor_asset:
                raise Exception(f"Invalid cursor string {cursor}")

            cursor_asset_str = cursor_asset.to_string()

            idx = len(asset_keys)
            for i in range(len(asset_keys)):
                if asset_keys[i].to_string() > cursor_asset_str:
                    idx = i
                    break

            asset_keys = asset_keys[idx:]

        if limit:
            asset_keys = asset_keys[:limit]
        return asset_keys

    @abstractmethod
    def get_latest_materialization_events(
        self, asset_keys: Iterable[AssetKey]
    ) -> Mapping[AssetKey, Optional["EventLogEntry"]]:
        pass

    def supports_add_asset_event_tags(self) -> bool:
        return False

    def add_asset_event_tags(
        self,
        event_id: int,
        event_timestamp: float,
        asset_key: AssetKey,
        new_tags: Mapping[str, str],
    ) -> None:
        raise NotImplementedError()

    @abstractmethod
    def get_event_tags_for_asset(
        self,
        asset_key: AssetKey,
        filter_tags: Optional[Mapping[str, str]] = None,
        filter_event_id: Optional[int] = None,
    ) -> Sequence[Mapping[str, str]]:
        pass

    def get_asset_tags_to_index(self, tag_keys: set[str]) -> set[str]:
        # make sure we update the list of tested tags in test_asset_tags_to_insert to match
        return {
            key
            for key in tag_keys
            if key == DATA_VERSION_TAG or key.startswith(MULTIDIMENSIONAL_PARTITION_PREFIX)
        }

    @abstractmethod
    def wipe_asset(self, asset_key: AssetKey) -> None:
        """Remove asset index history from event log for given asset_key."""

    @abstractmethod
    def wipe_asset_partitions(self, asset_key: AssetKey, partition_keys: Sequence[str]) -> None:
        """Remove asset index history from event log for given asset partitions."""

    @abstractmethod
    def get_materialized_partitions(
        self,
        asset_key: AssetKey,
        before_cursor: Optional[int] = None,
        after_cursor: Optional[int] = None,
    ) -> set[str]:
        pass

    @abstractmethod
    def get_latest_storage_id_by_partition(
        self,
        asset_key: AssetKey,
        event_type: DagsterEventType,
        partitions: Optional[set[str]] = None,
    ) -> Mapping[str, int]:
        pass

    @abstractmethod
    def get_latest_tags_by_partition(
        self,
        asset_key: AssetKey,
        event_type: DagsterEventType,
        tag_keys: Sequence[str],
        asset_partitions: Optional[Sequence[str]] = None,
        before_cursor: Optional[int] = None,
        after_cursor: Optional[int] = None,
    ) -> Mapping[str, Mapping[str, str]]:
        pass

    @abstractmethod
    def get_latest_asset_partition_materialization_attempts_without_materializations(
        self, asset_key: AssetKey, after_storage_id: Optional[int] = None
    ) -> Mapping[str, tuple[str, int]]:
        pass

    @abstractmethod
    def get_dynamic_partitions(self, partitions_def_name: str) -> Sequence[str]:
        """Get the list of partition keys for a dynamic partitions definition."""
        raise NotImplementedError()

    @abstractmethod
    def has_dynamic_partition(self, partitions_def_name: str, partition_key: str) -> bool:
        """Check if a dynamic partition exists."""
        raise NotImplementedError()

    @abstractmethod
    def add_dynamic_partitions(
        self, partitions_def_name: str, partition_keys: Sequence[str]
    ) -> None:
        """Add a partition for the specified dynamic partitions definition."""
        raise NotImplementedError()

    @abstractmethod
    def delete_dynamic_partition(self, partitions_def_name: str, partition_key: str) -> None:
        """Delete a partition for the specified dynamic partitions definition."""
        raise NotImplementedError()

    def alembic_version(self) -> Optional[AlembicVersion]:
        return None

    @property
    def is_run_sharded(self) -> bool:
        """Indicates that the EventLogStoarge is sharded."""
        return False

    @property
    def supports_global_concurrency_limits(self) -> bool:
        """Indicates that the EventLogStorage supports global concurrency limits."""
        return False

    @property
    def asset_records_have_last_observation(self) -> bool:
        return False

    @abstractmethod
    def initialize_concurrency_limit_to_default(self, concurrency_key: str) -> bool:
        """Initialize a concurrency limit to the instance default value.  Is a no-op for concurrency
        keys that are already configured with a limit.
        """
        raise NotImplementedError()

    @abstractmethod
    def set_concurrency_slots(self, concurrency_key: str, num: int) -> None:
        """Allocate concurrency slots for the given concurrency key."""
        raise NotImplementedError()

    @abstractmethod
    def delete_concurrency_limit(self, concurrency_key: str) -> None:
        """Delete concurrency limits and slots for the given concurrency key."""
        raise NotImplementedError()

    @abstractmethod
    def get_concurrency_keys(self) -> set[str]:
        """Get the set of concurrency limited keys."""
        raise NotImplementedError()

    @abstractmethod
    def get_concurrency_info(self, concurrency_key: str) -> ConcurrencyKeyInfo:
        """Get concurrency info for key."""
        raise NotImplementedError()

    @abstractmethod
    def get_pool_limits(self) -> Sequence[PoolLimit]:
        """Get the set of concurrency limited keys and limits."""
        raise NotImplementedError()

    @abstractmethod
    def claim_concurrency_slot(
        self,
        concurrency_key: str,
        run_id: str,
        step_key: str,
        priority: Optional[int] = None,
    ) -> ConcurrencyClaimStatus:
        """Claim concurrency slots for step."""
        raise NotImplementedError()

    @abstractmethod
    def check_concurrency_claim(
        self, concurrency_key: str, run_id: str, step_key: str
    ) -> ConcurrencyClaimStatus:
        """Claim concurrency slots for step."""
        raise NotImplementedError()

    @abstractmethod
    def get_concurrency_run_ids(self) -> set[str]:
        """Get a list of run_ids that are occupying or waiting for a concurrency key slot."""
        raise NotImplementedError()

    @abstractmethod
    def free_concurrency_slots_for_run(self, run_id: str) -> None:
        """Frees concurrency slots for a given run."""
        raise NotImplementedError()

    @abstractmethod
    def free_concurrency_slot_for_step(self, run_id: str, step_key: str) -> None:
        """Frees concurrency slots for a given run/step."""
        raise NotImplementedError()

    @property
    def supports_asset_checks(self):
        return True

    @abstractmethod
    def get_asset_check_execution_history(
        self,
        check_key: AssetCheckKey,
        limit: int,
        cursor: Optional[int] = None,
        status: Optional[AbstractSet[AssetCheckExecutionRecordStatus]] = None,
    ) -> Sequence[AssetCheckExecutionRecord]:
        """Get executions for one asset check, sorted by recency."""
        pass

    @abstractmethod
    def get_latest_asset_check_execution_by_key(
        self, check_keys: Sequence[AssetCheckKey]
    ) -> Mapping[AssetCheckKey, AssetCheckExecutionRecord]:
        """Get the latest executions for a list of asset checks."""
        pass

    @abstractmethod
    def fetch_materializations(
        self,
        records_filter: Union[AssetKey, AssetRecordsFilter],
        limit: int,
        cursor: Optional[str] = None,
        ascending: bool = False,
    ) -> EventRecordsResult:
        raise NotImplementedError()

    @abstractmethod
    def fetch_observations(
        self,
        records_filter: Union[AssetKey, AssetRecordsFilter],
        limit: int,
        cursor: Optional[str] = None,
        ascending: bool = False,
    ) -> EventRecordsResult:
        raise NotImplementedError()

    @property
    def supports_run_status_change_job_name_filter(self) -> bool:
        return False

    @abstractmethod
    def fetch_run_status_changes(
        self,
        records_filter: Union[DagsterEventType, RunStatusChangeRecordsFilter],
        limit: int,
        cursor: Optional[str] = None,
        ascending: bool = False,
    ) -> EventRecordsResult:
        raise NotImplementedError()

    @abstractmethod
    def get_latest_planned_materialization_info(
        self,
        asset_key: AssetKey,
        partition: Optional[str] = None,
    ) -> Optional[PlannedMaterializationInfo]:
        raise NotImplementedError()

    @abstractmethod
    def get_updated_data_version_partitions(
        self, asset_key: AssetKey, partitions: Iterable[str], since_storage_id: int
    ) -> set[str]:
        raise NotImplementedError()

    @property
    def handles_run_events_in_store_event(self) -> bool:
        return False

    def default_run_scoped_event_tailer_offset(self) -> int:
        return 0

    def get_asset_status_cache_values(
        self,
        partitions_defs_by_key: Iterable[tuple[AssetKey, Optional[PartitionsDefinition]]],
        context: LoadingContext,
    ) -> Sequence[Optional["AssetStatusCacheValue"]]:
        """Get the cached status information for each asset."""
        values = []
        for asset_key, partitions_def in partitions_defs_by_key:
            values.append(
                get_and_update_asset_status_cache_value(
                    self._instance, asset_key, partitions_def, loading_context=context
                )
            )
        return values

    def get_pool_config(self) -> PoolConfig:
        # Base implementation of fetching pool config.  To be overriden for remote storage
        # implementations where the local instance might not match the remote instance.
        return self._instance.get_concurrency_config().pool_config



# Copied from https://github.com/syrusakbary/aiodataloader

import sys
from asyncio import (
    AbstractEventLoop,
    Future,
    ensure_future,
    gather,
    get_event_loop,
    iscoroutine,
    iscoroutinefunction,
)
from collections import namedtuple
from collections.abc import Coroutine, Iterable, Iterator, MutableMapping
from functools import partial
from typing import Any, Callable, Generic, Optional, TypeVar, Union

if sys.version_info >= (3, 10):
    from typing import TypeGuard
else:
    from typing_extensions import TypeGuard

__version__ = "0.4.0"

KeyT = TypeVar("KeyT")
ReturnT = TypeVar("ReturnT")
CacheKeyT = TypeVar("CacheKeyT")
DataLoaderT = TypeVar("DataLoaderT", bound="DataLoader[Any, Any]")
T = TypeVar("T")


def iscoroutinefunctionorpartial(
    fn: Union[Callable[..., ReturnT], "partial[ReturnT]"],
) -> TypeGuard[Callable[..., Coroutine[Any, Any, ReturnT]]]:
    return iscoroutinefunction(fn.func if isinstance(fn, partial) else fn)


Loader = namedtuple("Loader", "key,future")


class _BaseDataLoader(Generic[KeyT, ReturnT]):
    def __init__(
        self,
        batch_load_fn: Callable[[Iterable[KeyT]], Coroutine[Any, Any, Iterable[ReturnT]]],
        max_batch_size: Optional[int] = None,
    ):
        self.max_batch_size = max_batch_size


class BlockingDataLoader(Generic[KeyT, ReturnT]):
    """Currently, the cache is not shared between blocking and non-blocking DataLoaders, as it is
    challenging to drive the event loop properly while managing a shared cache.
    """

    def __init__(
        self,
        batch_load_fn: Callable[[Iterable[KeyT]], Iterable[ReturnT]],
        get_cache_key: Optional[Callable[[KeyT], Union[CacheKeyT, KeyT]]] = None,
        max_batch_size: Optional[int] = None,
    ):
        self._cache = {}
        self._to_query = {}

        self.get_cache_key = get_cache_key or (lambda x: x)

        self.batch_load_fn = batch_load_fn
        self.max_batch_size = max_batch_size

    def prepare(self, keys: Iterable[KeyT]) -> None:
        # ensure that the provided keys will be fetched as a unit in the next fetch
        for key in keys:
            cache_key = self.get_cache_key(key)
            if cache_key not in self._cache:
                self._to_query[cache_key] = key

    def blocking_load(self, key: KeyT) -> ReturnT:
        """Loads the provided key synchronously, pulling from the cache if possible."""
        self.prepare([key])

        if self._to_query:
            for chunk in get_chunks(
                list(self._to_query.values()),
                self.max_batch_size or len(self._to_query),
            ):
                # uses independent event loop from the async system
                chunk_results = self.batch_load_fn(chunk)
                for k, v in zip(chunk, chunk_results):
                    self._cache[self.get_cache_key(k)] = v

        self._to_query = {}
        return self._cache[self.get_cache_key(key)]

    def blocking_load_many(self, keys: Iterable[KeyT]) -> Iterable[ReturnT]:
        self.prepare(keys)
        return [self.blocking_load(key) for key in keys]


class DataLoader(_BaseDataLoader[KeyT, ReturnT]):
    batch: bool = True
    max_batch_size: Optional[int] = None
    cache: Optional[bool] = True

    def __init__(
        self,
        batch_load_fn: Callable[[Iterable[KeyT]], Coroutine[Any, Any, Iterable[ReturnT]]],
        batch: Optional[bool] = None,
        max_batch_size: Optional[int] = None,
        cache: Optional[bool] = None,
        get_cache_key: Optional[Callable[[KeyT], Union[CacheKeyT, KeyT]]] = None,
        cache_map: Optional[MutableMapping[Union[CacheKeyT, KeyT], "Future[ReturnT]"]] = None,
        loop: Optional[AbstractEventLoop] = None,
    ):
        self._loop = loop

        self.batch_load_fn = batch_load_fn

        assert iscoroutinefunctionorpartial(
            self.batch_load_fn
        ), f"batch_load_fn must be coroutine. Received: {self.batch_load_fn}"

        if not callable(self.batch_load_fn):
            raise TypeError(
                "DataLoader must have a batch_load_fn which accepts "
                f"Iterable<key> and returns Future<Iterable<value>>, but got: {batch_load_fn}."
            )

        if batch is not None:
            self.batch = batch

        if max_batch_size is not None:
            self.max_batch_size = max_batch_size

        if cache is not None:
            self.cache = cache

        if get_cache_key is not None:
            self.get_cache_key = get_cache_key
        if not hasattr(self, "get_cache_key"):
            self.get_cache_key = lambda x: x

        self._cache = cache_map if cache_map is not None else {}
        self._queue: list[Loader] = []

        super().__init__(batch_load_fn, max_batch_size)

    @property
    def loop(self) -> AbstractEventLoop:
        # ensure a good error is thrown if the event loop has changed since instantiation,
        # can result in cryptic errors otherwise.

        # it might be possible to instead update the event loop if there is no work in flight,
        # unclear if there are valid uses cases for that

        if self._loop is None:
            # lazily initialize the loop to allow for cases where this object is constructed
            # outside the context of an event loop (e.g. unit tests)
            self._loop = get_event_loop()

        if self._loop != get_event_loop():
            raise Exception(
                "event loop has changed since init, this DataLoader instance is no longer usable."
            )

        return self._loop

    def load(self, key: KeyT) -> "Future[ReturnT]":
        """Loads a key, returning a `Future` for the value represented by that key."""
        if key is None:
            raise TypeError(
                "The loader.load() function must be called with a value, " f"but got: {key}."
            )

        cache_key = self.get_cache_key(key)

        # If caching and there is a cache-hit, return cached Future.
        if self.cache:
            cached_result = self._cache.get(cache_key)
            if cached_result:
                return cached_result

        # Otherwise, produce a new Future for this value.
        future = self.loop.create_future()
        # If caching, cache this Future.
        if self.cache:
            self._cache[cache_key] = future

        self.do_resolve_reject(key, future)
        return future

    def do_resolve_reject(self, key: KeyT, future: "Future[ReturnT]") -> None:
        # Enqueue this Future to be dispatched.
        self._queue.append(Loader(key=key, future=future))
        # Determine if a dispatch of this queue should be scheduled.
        # A single dispatch should be scheduled per queue at the time when the
        # queue changes from "empty" to "full".
        if len(self._queue) == 1:
            if self.batch:
                # If batching, schedule a task to dispatch the queue.
                enqueue_post_future_job(self.loop, self)
            else:
                # Otherwise dispatch the (queue of one) immediately.
                dispatch_queue(self)

    def load_many(self, keys: Iterable[KeyT]) -> "Future[list[ReturnT]]":
        """Loads multiple keys, returning a list of values.

        >>> a, b = await my_loader.load_many([ 'a', 'b' ])

        This is equivalent to the more verbose:

        >>> a, b = await gather(
        >>>    my_loader.load('a'),
        >>>    my_loader.load('b')
        >>> )
        """
        if not isinstance(keys, Iterable):
            raise TypeError(
                "The loader.load_many() function must be called with Iterable<key> "
                f"but got: {keys}."
            )

        return gather(*[self.load(key) for key in keys])

    def clear(self: DataLoaderT, key: KeyT) -> DataLoaderT:
        """Clears the value at `key` from the cache, if it exists. Returns itself for
        method chaining.
        """
        cache_key = self.get_cache_key(key)
        self._cache.pop(cache_key, None)
        return self

    def clear_all(self: DataLoaderT) -> DataLoaderT:
        """Clears the entire cache. To be used when some event results in unknown
        invalidations across this particular `DataLoader`. Returns itself for
        method chaining.
        """
        self._cache.clear()
        return self

    def prime(self: DataLoaderT, key: KeyT, value: ReturnT) -> DataLoaderT:
        """Adds the provied key and value to the cache. If the key already exists, no
        change is made. Returns itself for method chaining.
        """
        cache_key = self.get_cache_key(key)

        # Only add the key if it does not already exist.
        if cache_key not in self._cache:
            # Cache a rejected future if the value is an Error, in order to match
            # the behavior of load(key).
            future = self.loop.create_future()
            if isinstance(value, Exception):
                future.set_exception(value)
            else:
                future.set_result(value)

            self._cache[cache_key] = future

        return self


def enqueue_post_future_job(loop: AbstractEventLoop, loader: DataLoader[Any, Any]) -> None:
    async def dispatch() -> None:
        dispatch_queue(loader)

    loop.call_soon(ensure_future, dispatch())


def get_chunks(iterable_obj: list[T], chunk_size: int = 1) -> Iterator[list[T]]:
    chunk_size = max(1, chunk_size)
    return (iterable_obj[i : i + chunk_size] for i in range(0, len(iterable_obj), chunk_size))


def dispatch_queue(loader: DataLoader[Any, Any]) -> None:
    """Given the current state of a Loader instance, perform a batch load
    from its current queue.
    """
    # Take the current loader queue, replacing it with an empty queue.
    queue = loader._queue  # noqa: SLF001
    loader._queue = []  # noqa: SLF001

    # If a max_batch_size was provided and the queue is longer, then segment the
    # queue into multiple batches, otherwise treat the queue as a single batch.
    max_batch_size = loader.max_batch_size

    if max_batch_size and max_batch_size < len(queue):
        chunks = get_chunks(queue, max_batch_size)
        for chunk in chunks:
            # Note: the noqa here might mask a real bug and worthy of investigation
            ensure_future(dispatch_queue_batch(loader, chunk))  # noqa: RUF006
    else:
        ensure_future(dispatch_queue_batch(loader, queue))  # noqa: RUF006


async def dispatch_queue_batch(loader: DataLoader[Any, Any], queue: list[Loader]) -> None:
    # Collect all keys to be loaded in this dispatch
    keys = [ql.key for ql in queue]

    # Call the provided batch_load_fn for this loader with the loader queue's keys.
    try:
        batch_future = loader.batch_load_fn(keys)
    except Exception as e:
        return failed_dispatch(loader, queue, e)

    # Assert the expected response from batch_load_fn
    if not batch_future or not iscoroutine(batch_future):
        return failed_dispatch(
            loader,
            queue,
            TypeError(
                "DataLoader must be constructed with a function which accepts "
                "Iterable<key> and returns Future<Iterable<value>>, but the "
                f"function did not return a Coroutine: {batch_future}."
            ),
        )

    try:
        values = await batch_future
        if not isinstance(values, Iterable):
            raise TypeError(
                "DataLoader must be constructed with a function which accepts "
                "Iterable<key> and returns Future<Iterable<value>>, but the "
                f"function did not return a Future of a Iterable: {values}."
            )

        values = list(values)
        if len(values) != len(keys):
            raise TypeError(
                "DataLoader must be constructed with a function which accepts "
                "Iterable<key> and returns Future<Iterable<value>>, but the "
                "function did not return a Future of a Iterable with the same "
                "length as the Iterable of keys."
                f"\n\nKeys:\n{keys}"
                f"\n\nValues:\n{values}"
            )

        # Step through the values, resolving or rejecting each Future in the
        # loaded queue.
        for ql, value in zip(queue, values):
            if isinstance(value, Exception):
                ql.future.set_exception(value)
            else:
                ql.future.set_result(value)

    except Exception as e:
        return failed_dispatch(loader, queue, e)


def failed_dispatch(loader: DataLoader[Any, Any], queue: list[Loader], error: Exception) -> None:
    """Do not cache individual loads if the entire batch dispatch fails,
    but still reject each request so they do not hang.
    """
    for ql in queue:
        loader.clear(ql.key)
        ql.future.set_exception(error)
